"""Data generators."""

__author__ = ["Tveten"]

import numpy as np
import pandas as pd
import scipy.stats

from ..utils.validation.generation import check_random_generator, check_segment_lengths
from ._utils import recycle_list


def _check_distributions(
    distributions: (
        scipy.stats.rv_continuous
        | scipy.stats.rv_discrete
        | list[scipy.stats.rv_continuous]
        | list[scipy.stats.rv_discrete]
        | None
    ),
    n_segments: int,
) -> tuple[list[scipy.stats.rv_continuous | scipy.stats.rv_discrete], int, np.dtype]:
    """Check if distributions are valid and return as a list.

    Parameters
    ----------
    distributions : list of `scipy.stats.rv_continuous` or `scipy.stats.rv_discrete`
        List of distributions for each segment.

    n_segments : int
        Number of segments to generate. Used to check if the number of distributions
        matches the number of segments.

    Returns
    -------
    list[scipy.stats.rv_continuous | scipy.stats.rv_discrete]
        List of distributions for each segment, where each distribution is guarnteed
        to have a `rvs(size: int, random_state: int | None)` method that returns
        a numpy array or scalar of the same size, and the output size is either
        1 or `p`.
    int
        Output size of the distributions, which is either 1 or `p`.
    np.dtype
        Data type of the output of the distributions.
    """
    if distributions is None:
        distributions = [scipy.stats.norm(), scipy.stats.norm(5)]
    elif not isinstance(distributions, list):
        distributions = [distributions]

    if len(distributions) == 0:
        raise ValueError("distributions cannot be an empty list.")

    distributions = recycle_list(distributions, n_segments)

    output_sizes = []
    output_dtypes = []
    for dist in distributions:
        try:
            output = dist.rvs(size=1)
            output_sizes.append(output.size)
            output_dtypes.append(output.dtype)
        except Exception:
            output_sizes.append(None)

    if any(size is None for size in output_sizes):
        raise ValueError(
            "All distributions must support the 'rvs' method with a 'size' argument,"
            " where the output is a numpy.array or numpy scalar."
            " Ensure that all distributions are valid scipy.stats distributions."
        )

    if len(set(output_sizes)) > 1:
        raise ValueError(
            f"All distributions must produce samples with the same number of variables."
            f" Got distribution.rvs(size=1).size outputs: {output_sizes}."
        )

    if len(set(output_dtypes)) > 1:
        raise ValueError(
            "All distributions must produce samples with the same data type."
            f" Got distribution.rvs(size=1).dtype outputs: {output_dtypes}."
        )

    return distributions, output_sizes[0], output_dtypes[0]


def generate_piecewise_data(
    distributions: scipy.stats.rv_continuous
    | scipy.stats.rv_discrete
    | list[scipy.stats.rv_continuous]
    | list[scipy.stats.rv_discrete]
    | None = None,
    lengths: int | list[int] | np.ndarray | None = None,
    *,
    n_segments: int = 3,
    n_samples: int = 100,
    seed: int | np.random.Generator | None = None,
    return_params: bool = False,
) -> pd.DataFrame | tuple[pd.DataFrame, dict]:
    """Generate data with a piecewise constant distribution.

    Generate piecewise segments of data from `scipy.stats` distributions, where
    unspecified parameters are randomly generated.

    Parameters
    ----------
    distributions : list of `scipy.stats.rv_continuous` or `scipy.stats.rv_discrete`, optional (default=None)
        The distributions for generating piecewise data.
        They are recycled to match the number of segments specified by `lengths` or
        `n_segments`.
        If None, alternating segments of `scipy.stats.norm()` and `scipy.stats.norm(5)`
        are used. Each distribution is expected to be a scipy distribution instance
        (e.g., `scipy.stats.norm`, `scipy.stats.uniform`). See
        `scipy.stats <https://docs.scipy.org/doc/scipy/reference/stats.html>`_
        for a list of all available distributions.

    lengths : int, list of int or np.ndarray, optional (default=None)
        The segment lengths. There are three possible cases:

        1. `list` or `numpy array`: Custom set of segment lengths.
        2. `int`: Length of `n_segments` equal segments.
        3. `None`: Generate `n_segments` random segment lengths with a total sample size
           of `n_samples`.

    n_segments : int (default=3)
        Number of segments to generate if `lengths` is an integer or None.

    n_samples : int (default=100)
        Total number of samples to generate if `lengths` is not specified.

    seed : np.random.Generator | int | None, optional
        Seed for the random number generator or a numpy random generator instance.
        If specified, this ensures reproducible output across multiple calls.

    return_params : bool, optional (default=False)
        If True, the function returns a tuple of the generated DataFrame and a
        dictionary with the parameters used to generate the data.

    Returns
    -------
    pd.DataFrame
        Data frame with generated data.

    dict, optional
        A dictionary containing the parameters used to generate the data. Only returned
        if `return_params` is True. It has the following keys:

        * `"n_segments"` : number of segments generated.
        * `"n_samples"` : total number of samples generated.
        * `"distributions"` : list of `scipy.stats.rv_continuous` or
          `scipy.stats.rv_discrete` with the distributions used for each segment.
        * `"lengths"` : list of lengths for each segment.
        * `"change_points"` : list of change points, which are the starting indices
          of each segment in the data.

    Examples
    --------
    >>> # Example 1: Two normal segments
    >>> from skchange.datasets import generate_piecewise_data
    >>> from scipy.stats import norm
    >>> generate_piecewise_data(
    ...     distributions=[norm(0, 1), norm(10, 0.1)],
    ...     lengths=[7, 3],
    ...     seed=1,
    ... )
               0
    0   0.345584
    1   0.821618
    2   0.330437
    3  -1.303157
    4   0.905356
    5   0.446375
    6  -0.536953
    7  10.058112
    8  10.036457
    9  10.029413

    >>> # Example 2: Two Poisson segments
    >>> from scipy.stats import poisson
    >>> generate_piecewise_data(
    ...     distributions=[poisson(1), poisson(10)],
    ...     lengths=[5, 5],
    ...     seed=2,
    ... )
        0
    0   0
    1   0
    2   1
    3   2
    4   0
    5   8
    6  11
    7   9
    8   9
    9   9


    >>> # Example 3: Specify int lengths and n_segments
    >>> generate_piecewise_data(
    ...     distributions=[norm(0), norm(5)],
    ...     lengths=3,
    ...     n_segments=3,
    ...     seed=3,
    ... )
              0
    0  2.040919
    1 -2.555665
    2  0.418099
    3  4.432230
    4  4.547351
    5  4.784403
    6 -2.019986
    7 -0.231932
    8 -0.865213
    """  # noqa: E501
    random_generator = check_random_generator(seed)
    lengths = check_segment_lengths(
        lengths, n_segments, n_samples, seed=random_generator
    )
    n_segments = len(lengths)
    n_samples = np.sum(lengths)
    distributions, n_variables, dtype = _check_distributions(distributions, n_segments)

    ends = np.cumsum(lengths)
    starts = np.concatenate(([0], ends[:-1]))
    generated_values = np.empty((n_samples, n_variables), dtype=dtype)
    for distribution, start, end in zip(distributions, starts, ends):
        length = end - start
        values = distribution.rvs(size=length, random_state=random_generator)
        generated_values[start:end, :] = values.reshape(length, n_variables)

    generated_df = pd.DataFrame(generated_values)

    if return_params:
        params = {
            "n_segments": n_segments,
            "n_samples": n_samples,
            "lengths": lengths,
            "distributions": distributions,
            "change_points": starts[1:],
        }
        return generated_df, params

    return generated_df



"""Base classes for anomaly detectors.

    classes:
        BaseSegmentAnomalyDetector

By inheriting from these classes the remaining methods of the BaseDetector class to
implement to obtain a fully functional anomaly detector are given below.

Needs to be implemented:
    _fit(self, X, y=None)
    _predict(self, X)

Optional to implement:
    _transform_scores(self, X)
    _update(self, X, y=None)
"""

import numpy as np
import pandas as pd

from ..base import BaseDetector


class BaseSegmentAnomalyDetector(BaseDetector):
    """Base class for segment anomaly detectors.

    Segment anomaly detectors detect segments of data points that are considered
    anomalous.

    Output format of the `predict` method: See the `dense_to_sparse` method.
    Output format of the `transform` method: See the `sparse_to_dense` method.
    """

    _tags = {
        "authors": ["Tveten"],
        "maintainers": ["Tveten"],
        "task": "segmentation",
    }

    @staticmethod
    def sparse_to_dense(
        y_sparse: pd.DataFrame, index: pd.Index, columns: pd.Index = None
    ) -> pd.DataFrame:
        """Convert the sparse output from the `predict` method to a dense format.

        Parameters
        ----------
        y_sparse : pd.DataFrame with RangeIndex
            Detected segment anomalies. Must have the following column:

            * ``"ilocs"`` -  left-closed intervals of iloc based segments.

            Can also have the following columns:

            * ``"icolumns"`` - array of identified variables for each anomaly.
        index : array-like
            Indices that are to be annotated according to `y_sparse`.
        columns : array-like
            Columns that are to be annotated according to `y_sparse`. Only relevant if
            y_sparse contains the column ``"icolumns"`` with identified variables.

        Returns
        -------
        pd.DataFrame with the input data index and one column:
            * ``"label"`` - integer labels ``1, ..., K`` for each segment anomaly.
            ``0`` is reserved for the normal instances.
        """
        if "icolumns" in y_sparse:
            return BaseSegmentAnomalyDetector._sparse_to_dense_icolumns(
                y_sparse, index, columns
            )
        return BaseSegmentAnomalyDetector._sparse_to_dense_ilocs(y_sparse, index)

    @staticmethod
    def dense_to_sparse(y_dense: pd.DataFrame) -> pd.DataFrame:
        """Convert the dense output from the `transform` method to a sparse format.

        Parameters
        ----------
        y_dense : pd.DataFrame
            The dense output from the `transform` method. It must either have the
            following column:

            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly and
            label 0 for normal instances.

            Or it must have columns of the form:

            * ``"labels_<*>"`` with integer labels ``1, ..., K`` for each segment
            anomaly, and 0 for normal instances.

        Returns
        -------
        pd.DataFrame :
            A ``pd.DataFrame`` with a range index and two columns:
            * ``"ilocs"`` - left-closed `pd.Interval`s of iloc based segments.
            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly.

        Notes
        -----
        The start and end points of the intervals can be accessed by
        ``output["ilocs"].array.left`` and ``output["ilocs"].array.right``,
        respectively.
        """
        if "labels" in y_dense.columns:
            return BaseSegmentAnomalyDetector._dense_to_sparse_ilocs(y_dense)
        elif y_dense.columns.str.startswith("labels_").all():
            return BaseSegmentAnomalyDetector._dense_to_sparse_icolumns(y_dense)
        raise ValueError(
            "Invalid columns in `y_dense`. Expected 'labels' or 'labels_*'."
            f" Got: {y_dense.columns}"
        )

    def _format_sparse_output(
        self,
        segment_anomalies: list[tuple[int, int]] | list[tuple[int, int, np.ndarray]],
        closed: str = "left",
    ) -> pd.DataFrame:
        """Format the sparse output of segment anomaly detectors.

        Can be reused by subclasses to format the output of the `_predict` method.

        Parameters
        ----------
        segment_anomalies : list
            List of tuples containing start and end indices of segment anomalies,
            and optionally a ``np.ndarray`` of the identified
            variables/components/columns.
        closed : str
            Whether the ``(start, end)`` tuple correspond to intervals that are closed
            on the left, right, both, or neither.

        Returns
        -------
        pd.DataFrame :
            A ``pd.DataFrame`` with a range index and two columns:
            * ``"ilocs"`` - left-closed ``pd.Interval``s of iloc based segments.
            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly.

        Notes
        -----
        The start and end points of the intervals can be accessed by
        ``output["ilocs"].array.left`` and ``output["ilocs"].array.right``,
        respectively.
        """
        # Cannot extract this from segment_anomalies as it may be an empty list.
        if self.get_tag("capability:variable_identification"):
            return self._format_sparse_output_icolumns(segment_anomalies, closed)
        else:
            return self._format_sparse_output_ilocs(segment_anomalies, closed)

    @staticmethod
    def _sparse_to_dense_ilocs(
        y_sparse: pd.DataFrame, index: pd.Index, columns: pd.Index = None
    ) -> pd.DataFrame:
        """Convert the sparse output from the `predict` method to a dense format.

        Parameters
        ----------
        y_sparse : pd.DataFrame with RangeIndex
            Detected segment anomalies. Must have the following column:

            * ``"ilocs"`` -  left-closed intervals of iloc based segments.
        index : array-like
            Indices that are to be annotated according to `y_sparse`.
        columns: array-like
            Not used. Only for API compatibility.

        Returns
        -------
        pd.DataFrame with the input data index and one column:
            * ``"label"`` - integer labels ``1, ..., K`` for each segment anomaly.
            ``0`` is reserved for the normal instances.
        """
        labels = pd.IntervalIndex(y_sparse["ilocs"]).get_indexer(index)
        # `get_indexer` return values 0 for the values inside the first interval, 1 to
        # the values within the next interval and so on, and -1 for values outside any
        # interval. The `skchange` convention is that 0 is normal and > 0 is anomalous,
        # so we add 1 to the result.
        labels += 1
        return pd.DataFrame(labels, index=index, columns=["labels"], dtype="int64")

    @staticmethod
    def _dense_to_sparse_ilocs(y_dense: pd.DataFrame) -> pd.DataFrame:
        """Convert the dense output from the `transform` method to a sparse format.

        Parameters
        ----------
        y_dense : pd.DataFrame
            The dense output from the `transform` method. Must have the following
            column:

            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly and
            label ``0`` for normal instances.

        Returns
        -------
        pd.DataFrame :
            A `pd.DataFrame` with a range index and two columns:
            * ``"ilocs"`` - left-closed ``pd.Interval``s of iloc based segments.
            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly.

        Notes
        -----
        The start and end points of the intervals can be accessed by
        ``output["ilocs"].array.left`` and ``output["ilocs"].array.right``,
        respectively.
        """
        # The sparse format only uses integer positions, so we reset the index.
        y_dense = y_dense["labels"].reset_index(drop=True)

        y_anomaly = y_dense.loc[y_dense.values > 0]
        anomaly_locations_diff = y_anomaly.index.diff()

        first_anomaly_start = y_anomaly.index[:1].to_numpy()
        anomaly_starts = y_anomaly.index[anomaly_locations_diff > 1]
        anomaly_starts = np.insert(anomaly_starts, 0, first_anomaly_start)

        last_anomaly_end = y_anomaly.index[-1:].to_numpy() + 1
        anomaly_ends = y_anomaly.index[np.roll(anomaly_locations_diff > 1, -1)] + 1
        anomaly_ends = np.insert(anomaly_ends, len(anomaly_ends), last_anomaly_end)

        anomaly_intervals = list(zip(anomaly_starts, anomaly_ends))
        return BaseSegmentAnomalyDetector._format_sparse_output_ilocs(
            anomaly_intervals, closed="left"
        )

    @staticmethod
    def _format_sparse_output_ilocs(
        anomaly_intervals: list[tuple[int, int]], closed: str = "left"
    ) -> pd.DataFrame:
        """Format the sparse output of segment anomaly detectors.

        Can be reused by subclasses to format the output of the `_predict` method.

        Parameters
        ----------
        anomaly_intervals : list
            List of tuples containing start and end indices of segment anomalies.

        Returns
        -------
        pd.DataFrame :
            A `pd.DataFrame` with a range index and two columns:
            * ``"ilocs"`` - left-closed ``pd.Interval``s of iloc based segments.
            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly.

        Notes
        -----
        The start and end points of the intervals can be accessed by
        ``output["ilocs"].array.left`` and ``output["ilocs"].array.right``,
        respectively.
        """
        anomaly_intervals = [(int(start), int(end)) for start, end in anomaly_intervals]
        return pd.DataFrame(
            {
                "ilocs": pd.IntervalIndex.from_tuples(anomaly_intervals, closed=closed),
                "labels": pd.RangeIndex(1, len(anomaly_intervals) + 1),
            },
        )

    @staticmethod
    def _sparse_to_dense_icolumns(
        y_sparse: pd.DataFrame, index: pd.Index, columns: pd.Index
    ) -> pd.DataFrame:
        """Convert the sparse output from the `predict` method to a dense format.

        Parameters
        ----------
        y_sparse : pd.DataFrame with RangeIndex
            Detected segment anomalies. Must have the following columns:

            * ``"ilocs"`` -  left-closed intervals of iloc based segments.
            * ``"icolumns"`` - array of identified variables for each anomaly.
        index : array-like
            Indices that are to be annotated according to `y_sparse`.
        columns : array-like
            Columns that are to be annotated according to `y_sparse`.

        Returns
        -------
        pd.DataFrame with the input data index and as many columns as in X:
            * ``"labels_<X.columns[i]>"`` for each column index i in ``X.columns``:
            Integer labels starting from ``0``.
        """
        anomaly_intervals = y_sparse["ilocs"].array
        anomaly_starts = anomaly_intervals.left
        anomaly_ends = anomaly_intervals.right
        anomaly_columns = y_sparse["icolumns"]

        start_is_open = anomaly_intervals.closed in ["neither", "right"]
        if start_is_open:
            anomaly_starts += 1  # Exclude the start index in the for loop below.
        end_is_closed = anomaly_intervals.closed in ["both", "right"]
        if end_is_closed:
            anomaly_ends += 1  # Include the end index in the for loop below.

        labels = np.zeros((len(index), len(columns)), dtype="int64")
        anomalies = zip(anomaly_starts, anomaly_ends, anomaly_columns)
        for i, (start, end, affected_columns) in enumerate(anomalies):
            labels[start:end, affected_columns] = i + 1

        prefixed_columns = [f"labels_{column}" for column in columns]
        return pd.DataFrame(labels, index=index, columns=prefixed_columns)

    @staticmethod
    def _dense_to_sparse_icolumns(y_dense: pd.DataFrame):
        """Convert the dense output from the `transform` method to a sparse format.

        Parameters
        ----------
        y_dense : pd.DataFrame
            The dense output from the `transform` method. Must have columns of the form:

            * `"labels_<*>"` with integer labels ``1, ..., K`` for each segment anomaly,
            and ``0`` for normal instances.

        Returns
        -------
        pd.DataFrame :
            A ``pd.DataFrame`` with a range index and three columns:
            * ``"ilocs"`` - left-closed ``pd.Interval``s of iloc based segments.
            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly.
            * ``"icolumns"`` - list of affected columns for each anomaly.
        """
        # The sparse format only uses integer positions, so we reset index and columns.
        y_dense = y_dense.reset_index(drop=True)
        y_dense.columns = range(y_dense.columns.size)

        anomaly_intervals = []
        unique_labels = np.unique(y_dense.values)
        for i in unique_labels[unique_labels > 0]:
            anomaly_mask = y_dense == i
            which_columns = anomaly_mask.any(axis=0)
            which_rows = anomaly_mask.any(axis=1)
            anomaly_columns = anomaly_mask.columns[which_columns].to_list()
            anomaly_start = anomaly_mask.index[which_rows][0]
            anomaly_end = anomaly_mask.index[which_rows][-1]
            anomaly_intervals.append((anomaly_start, anomaly_end + 1, anomaly_columns))

        return BaseSegmentAnomalyDetector._format_sparse_output_icolumns(
            anomaly_intervals, closed="left"
        )

    @staticmethod
    def _format_sparse_output_icolumns(
        segment_anomalies: list[tuple[int, int, np.ndarray]],
        closed: str = "left",
    ) -> pd.DataFrame:
        """Format the sparse output of subset segment anomaly detectors.

        Can be reused by subclasses to format the output of the `_predict` method.

        Parameters
        ----------
        segment_anomalies : list
            List of tuples containing start and end indices of segment
            anomalies and a ``np.array`` of the affected components/columns.
        closed : str
            Whether the ``(start, end)`` tuple correspond to intervals that are closed
            on the left, right, both, or neither.

        Returns
        -------
        pd.DataFrame :
            A ``pd.DataFrame`` with a range index and three columns:
            * ``"ilocs"`` - left-closed ``pd.Interval``s of iloc based segments.
            * ``"labels"`` - integer labels ``1, ..., K`` for each segment anomaly.
            * ``"icolumns"`` - list of affected columns for each anomaly.
        """
        ilocs = [(int(start), int(end)) for start, end, _ in segment_anomalies]
        icolumns = [
            np.array(components, dtype="int64")
            for _, _, components in segment_anomalies
        ]
        return pd.DataFrame(
            {
                "ilocs": pd.IntervalIndex.from_tuples(ilocs, closed=closed),
                "labels": pd.RangeIndex(1, len(ilocs) + 1),
                "icolumns": icolumns,
            }
        )



"""Interactively compare the outputs of the different detector types."""

import numpy as np
import plotly.express as px

from skchange.anomaly_detectors import CAPA
from skchange.anomaly_scores import L2Saving
from skchange.change_detectors import MovingWindow
from skchange.datasets import generate_anomalous_data
from skchange.penalties import make_linear_chi2_penalty

# Generate data
n = 300
means = [np.array([8.0, 0.0, 0.0]), np.array([2.0, 3.0, 5.0])]
df = generate_anomalous_data(
    n, anomalies=[(100, 120), (250, 300)], means=means, random_state=3
)
df.columns += 1
plot_df = (
    df.stack()
    .reset_index()
    .rename({"level_0": "time", "level_1": "variable", 0: "value"}, axis=1)
)
fig = px.line(plot_df, x="time", y="value", facet_row="variable")
fig.show()

# Change detector
change_detector = MovingWindow(threshold_scale=1.0)
changepoints = change_detector.fit_predict(df)
changepoint_labels = change_detector.transform(df)
print(changepoints)
print(changepoint_labels)

# Segment anomaly detector
anomaly_detector = CAPA()
anomalies = anomaly_detector.fit_predict(df)
anomaly_labels = anomaly_detector.transform(df)
print(anomalies)
print(anomaly_labels)

# Subset segment anomaly detector
score = L2Saving()
p = df.shape[1]
penalty = make_linear_chi2_penalty(score.get_model_size(p), n, p)
subset_anomaly_detector = CAPA(score, penalty, find_affected_components=True)
subset_anomalies = subset_anomaly_detector.fit_predict(df)
subset_anomaly_labels = subset_anomaly_detector.transform(df)
print(subset_anomalies)
print(subset_anomaly_labels)



import numpy as np
import pytest

from skchange.anomaly_scores import SAVINGS, to_saving
from skchange.compose.penalised_score import PenalisedScore
from skchange.costs import COSTS
from skchange.costs.tests.test_all_costs import create_fixed_cost_test_instance
from skchange.datasets import generate_alternating_data
from skchange.tests.test_all_interval_scorers import skip_if_no_test_data

COST_INSTANCES = [
    to_saving(create_fixed_cost_test_instance(cost))
    for cost in COSTS
    if cost.create_test_instance().get_tag("supports_fixed_param")
]
SAVING_INSTANCES = [saving.create_test_instance() for saving in SAVINGS]
PENALISED_SAVING_INSTANCES = [
    PenalisedScore(saving)
    for saving in COST_INSTANCES + SAVING_INSTANCES
    if not saving.get_tag("is_penalised")
]
ALL_SAVING_INSTANCES = COST_INSTANCES + SAVING_INSTANCES + PENALISED_SAVING_INSTANCES


@pytest.mark.parametrize("saving", ALL_SAVING_INSTANCES)
def test_saving_values(saving):
    """Test all available savings."""
    skip_if_no_test_data(saving)

    n = 50
    df = generate_alternating_data(n_segments=1, segment_length=n, p=1, random_state=5)
    saving.fit(df)

    starts = np.arange(n - 15)
    ends = np.repeat(n - 1, len(starts))
    intervals = np.column_stack((starts, ends))
    saving_values = saving.evaluate(intervals)

    saving_is_penalised = saving.get_tag("is_penalised")
    min_value = 0.0 if not saving_is_penalised else -np.inf
    assert np.all(saving_values >= min_value)



"""Interactive exploration of the MovingWindow change detector."""

import plotly.express as px

from skchange.change_detectors._moving_window import MovingWindow
from skchange.datasets import generate_alternating_data
from skchange.utils.benchmarking.profiler import Profiler

# Simple univariate example
df = generate_alternating_data(
    n_segments=2, mean=10, segment_length=100, p=1, random_state=2
)
detector = MovingWindow()
changepoints = detector.fit_predict(df)
labels = detector.transform(df)
scores = detector.transform_scores(df)
px.scatter(scores)


# Profiling
n = int(1e6)
df = generate_alternating_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = MovingWindow(bandwidth=50)
profiler = Profiler()
profiler.start()
detector.fit_predict(df)
profiler.stop()


# Multiple bandwidths
df = generate_alternating_data(
    n_segments=2, mean=10, segment_length=100, p=1, random_state=2
)
detector = MovingWindow(bandwidth=[10, 20, 30, 40, 50, 60, 100])
changepoints = detector.fit_predict(df)
labels = detector.transform(df)
scores = detector.transform_scores(df)
px.scatter(scores)



"""Seeded binary segmentation algorithm for multiple changepoint detection."""

__author__ = ["Tveten"]
__all__ = ["SeededBinarySegmentation"]

import numpy as np
import pandas as pd

from ..base import BaseIntervalScorer
from ..change_scores import CUSUM, to_change_score
from ..compose.penalised_score import PenalisedScore
from ..utils.numba import njit
from ..utils.validation.data import check_data
from ..utils.validation.interval_scorer import check_interval_scorer
from ..utils.validation.parameters import check_in_interval, check_larger_than
from ..utils.validation.penalties import check_penalty
from .base import BaseChangeDetector


@njit
def make_seeded_intervals(
    n: int, min_length: int, max_length: int, growth_factor: float = 1.5
) -> tuple[np.ndarray, np.ndarray]:
    starts = [0]  # For numba to be able to compile type.
    ends = [1]  # For numba to be able to compile type.
    step_factor = 1 - 1 / growth_factor
    max_length = min(max_length, n)
    n_lengths = int(np.ceil(np.log(max_length / min_length) / np.log(growth_factor)))
    interval_lens = np.unique(np.round(np.geomspace(min_length, max_length, n_lengths)))
    for interval_len in interval_lens:
        step = max(1, np.round(step_factor * interval_len))
        n_steps = int(np.ceil((n - interval_len) / step))
        new_starts = [int(i * step) for i in range(n_steps + 1)]
        starts += new_starts
        new_ends = [int(min(i * step + interval_len, n)) for i in range(n_steps + 1)]
        ends += new_ends
        if ends[-1] - starts[-1] < min_length:
            starts[-1] = n - min_length
    return np.array(starts[1:]), np.array(ends[1:])


@njit
def greedy_selection(
    max_scores: np.ndarray,
    argmax_scores: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
) -> list[int]:
    max_scores = max_scores.copy()
    cpts = []
    while np.any(max_scores > 0):
        argmax = max_scores.argmax()
        cpt = argmax_scores[argmax]
        cpts.append(int(cpt))
        # remove intervals that contain the detected changepoint.
        max_scores[(cpt >= starts) & (cpt < ends)] = 0.0
    cpts.sort()
    return cpts


@njit
def narrowest_selection(
    max_scores: np.ndarray,
    argmax_scores: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
) -> list[int]:
    cpts = []
    scores_above_threshold = max_scores > 0
    candidate_starts = starts[scores_above_threshold]
    candidate_ends = ends[scores_above_threshold]
    candidate_maximizers = argmax_scores[scores_above_threshold]

    while len(candidate_starts) > 0:
        argmin = np.argmin(candidate_ends - candidate_starts)
        cpt = candidate_maximizers[argmin]
        cpts.append(int(cpt))

        # remove candidates that contain the detected changepoint.
        cpt_not_in_interval = ~((cpt >= candidate_starts) & (cpt < candidate_ends))
        candidate_starts = candidate_starts[cpt_not_in_interval]
        candidate_ends = candidate_ends[cpt_not_in_interval]
        candidate_maximizers = candidate_maximizers[cpt_not_in_interval]

    cpts.sort()
    return cpts


def run_seeded_binseg(
    penalised_score: BaseIntervalScorer,
    max_interval_length: int,
    growth_factor: float,
    selection_method: str = "greedy",
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    penalised_score.check_is_penalised()
    penalised_score.check_is_fitted()
    n_samples = penalised_score.n_samples

    starts, ends = make_seeded_intervals(
        n_samples,
        2 * penalised_score.min_size,
        max_interval_length,
        growth_factor,
    )

    max_scores = np.zeros(starts.size)
    argmax_scores = np.zeros(starts.size, dtype=np.int64)
    for i, (start, end) in enumerate(zip(starts, ends)):
        splits = np.arange(
            start + penalised_score.min_size, end - penalised_score.min_size + 1
        )
        intervals = np.column_stack(
            (np.repeat(start, splits.size), splits, np.repeat(end, splits.size))
        )
        scores = penalised_score.evaluate(intervals)
        argmax = np.argmax(scores)
        max_scores[i] = scores[argmax, 0]  # index 0 to get a scalar value
        argmax_scores[i] = splits[0] + argmax

    if selection_method == "greedy":
        cpts = greedy_selection(max_scores, argmax_scores, starts, ends)
    elif selection_method == "narrowest":
        cpts = narrowest_selection(max_scores, argmax_scores, starts, ends)

    return cpts, max_scores, argmax_scores, starts, ends


class SeededBinarySegmentation(BaseChangeDetector):
    """Seeded binary segmentation algorithm for multiple changepoint detection.

    Binary segmentation type changepoint detection algorithms recursively split the data
    into two segments, and test whether the two segments are different. The seeded
    binary segmentation algorithm is an efficient version of such algorithms, which
    tests for changepoints in intervals of exponentially growing length. It has the same
    theoretical guarantees as the original binary segmentation algorithm, but runs
    in log-linear time no matter the changepoint configuration.

    Parameters
    ----------
    change_score : BaseIntervalScorer, optional, default=CUSUM()
        The change score to use in the algorithm. If a cost is given, it is
        converted to a change score using the `ChangeScore` class.
    penalty : np.ndarray or float, optional, default=None
        The penalty to use for change detection. If the penalty is
        penalised (`change_score.get_tag("is_penalised")`) the penalty will
        be ignored. The different types of penalties are as follows:

        * ``float``: A constant penalty applied to the sum of scores across all
          variables in the data.
        * ``np.ndarray``: A penalty array of the same length as the number of
          columns in the data, where element ``i`` of the array is the penalty for
          ``i+1`` variables being affected by a change. The penalty array
          must be positive and increasing (not strictly). A penalised score with a
          linear penalty array is faster to evaluate than a nonlinear penalty array.
        * ``None``: A default penalty is created in `predict` based on the fitted
          score using the `make_bic_penalty` function.

    max_interval_length : int, default=200
        The maximum length of an interval to estimate a changepoint in. Must be greater
        than or equal to ``2 * change_score.min_size``.
    growth_factor : float, default=1.5
        The growth factor for the seeded intervals. Intervals grow in size according to
        ``interval_len=max(interval_len + 1, np.floor(growth_factor * interval_len))``,
        starting at ``interval_len=min_interval_length``. It also governs the amount
        of overlap between intervals of the same length, as the start of each interval
        is shifted by a factor of ``1 + 1 / growth_factor``. Must be a float in
        ``(1, 2]``.
    selection_method : str, default="greedy"
        The method to use for selecting changepoints. The options are:

        * ``"greedy"``: Selects the changepoint with the highest score, then removes all
          intervals that contain the detected changepoint. This process is repeated
          until no intervals are left with a score above the threshold.
        * ``"narrowest"``: Searches among the intervals with scores above the threshold,
          and selects the one with the narrowest interval. It then removes all
          intervals that contain the detected changepoint, and repeats these two steps
          until no intervals are left with a score above the threshold.

    References
    ----------
    .. [1] Kovács, S., Bühlmann, P., Li, H., & Munk, A. (2023). Seeded binary
        segmentation: a general methodology for fast and optimal changepoint detection.
        Biometrika, 110(1), 249-256.

    .. [2] Baranowski, R., Chen, Y., & Fryzlewicz, P. (2019). Narrowest-over-threshold
        detection of multiple change points and change-point-like features. Journal of
        the Royal Statistical Society Series B: Statistical Methodology, 81(3), 649-672.

    Examples
    --------
    >>> from skchange.change_detectors import SeededBinarySegmentation
    >>> from skchange.datasets import generate_alternating_data
    >>> df = generate_alternating_data(n_segments=4, mean=10, segment_length=100, p=5)
    >>> detector = SeededBinarySegmentation()
    >>> detector.fit_predict(df)
       ilocs
    0    100
    1    200
    2    300
    """

    _tags = {
        "authors": ["Tveten"],
        "maintainers": ["Tveten"],
        "fit_is_empty": True,
    }

    def __init__(
        self,
        change_score: BaseIntervalScorer | None = None,
        penalty: np.ndarray | float | None = None,
        max_interval_length: int = 200,
        growth_factor: float = 1.5,
        selection_method: str = "greedy",
    ):
        self.change_score = change_score
        self.penalty = penalty
        self.max_interval_length = max_interval_length
        self.growth_factor = growth_factor
        self.selection_method = selection_method
        super().__init__()

        _score = CUSUM() if change_score is None else change_score
        check_interval_scorer(
            _score,
            "change_score",
            "SeededBinarySegmentation",
            required_tasks=["cost", "change_score"],
            allow_penalised=True,
        )
        _change_score = to_change_score(_score)

        check_penalty(penalty, "penalty", "SeededBinarySegmentation")
        self._penalised_score = (
            _change_score.clone()  # need to avoid modifying the input change_score
            if _change_score.get_tag("is_penalised")
            else PenalisedScore(_change_score, penalty)
        )

        check_in_interval(
            pd.Interval(1.0, 2.0, closed="right"),
            self.growth_factor,
            "growth_factor",
        )
        valid_selection_methods = ["greedy", "narrowest"]
        if self.selection_method not in valid_selection_methods:
            raise ValueError(
                f"Invalid selection method. Must be one of {valid_selection_methods}."
            )

        self.clone_tags(_change_score, ["distribution_type"])

    def _predict(self, X: pd.DataFrame | pd.Series) -> pd.Series:
        """Detect events in test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame
            Time series to detect change points in.

        Returns
        -------
        y_sparse : pd.DataFrame
            A `pd.DataFrame` with a range index and one column:
            * ``"ilocs"`` - integer locations of the change points.

        Attributes
        ----------
        fitted_score : BaseIntervalScorer
            The fitted penalised change score used for the detection.
        scores: pd.DataFrame
            A `pd.DataFrame` with the following columns:
            * ``"start"`` - start of the interval.
            * ``"end"`` - end of the interval.
            * ``"argmax"`` - index of the maximum score in the interval.
            * ``"max"`` - maximum score in the interval.
        """
        self.fitted_score: BaseIntervalScorer = self._penalised_score.clone()
        self.fitted_score.fit(X)
        X = check_data(
            X,
            min_length=2 * self.fitted_score.min_size,
            min_length_name="2 * fitted_change_score.min_size",
        )
        check_larger_than(
            2 * self.fitted_score.min_size,
            self.max_interval_length,
            "max_interval_length",
        )
        cpts, scores, maximizers, starts, ends = run_seeded_binseg(
            penalised_score=self.fitted_score,
            max_interval_length=self.max_interval_length,
            growth_factor=self.growth_factor,
            selection_method=self.selection_method,
        )
        self.scores = pd.DataFrame(
            {"start": starts, "end": ends, "argmax": maximizers, "max": scores}
        )
        return self._format_sparse_output(cpts)

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skchange.costs import L2Cost

        params = [
            {
                "change_score": L2Cost(),
                "max_interval_length": 100,
                "penalty": 30,
            },
            {
                "change_score": L2Cost(),
                "max_interval_length": 20,
                "penalty": 10,
            },
            {
                "change_score": L2Cost(),
                "max_interval_length": 20,
                "penalty": 10,
                "selection_method": "narrowest",
            },
        ]
        return params



"""Numba-optimized functions for various common data manipulation tasks."""

import numpy as np

from skchange.utils.numba import njit, prange


@njit
def col_repeat(x: np.ndarray, n: int) -> np.ndarray:
    """Repeat each column of a 2D array n times.

    Parameters
    ----------
    x : np.ndarray
        1D array.

    Returns
    -------
    2D array : (x.size, n)-matrix with x in each column
    """
    expanded_x = np.zeros((x.shape[0], n))
    for j in prange(n):
        expanded_x[:, j] = x
    return expanded_x


@njit
def row_repeat(x: np.ndarray, n: int) -> np.ndarray:
    """Repeat each row of a 2D array n times.

    Parameters
    ----------
    x : np.ndarray
        1D array.

    Returns
    -------
    2D array : (n, x.size) matrix with x in each row
    """
    expanded_x = np.zeros((x.shape[0], n))
    for i in prange(n):
        expanded_x[i, :] = x
    return expanded_x


@njit
def where(indicator: np.ndarray) -> list:
    """
    Identify consecutive intervals of True values in the input array.

    Parameters
    ----------
    indicator : np.ndarray
        1D boolean array.

    Returns
    -------
    list of tuples:
        Each tuple represents the start and end indices of consecutive True intervals.
        If there are no True values, an empty list is returned.
    """
    intervals = []
    start, end = None, None
    for i, val in enumerate(indicator):
        if val and start is None:
            start = i
        elif not val and start is not None:
            end = i
            intervals.append((start, end))
            start, end = None, None
    if start is not None:
        intervals.append((start, len(indicator)))
    return intervals


@njit
def truncate_below(x: np.ndarray, lower_bound: float) -> np.ndarray:
    """Truncate values below a lower bound.

    Parameters
    ----------
    x : np.ndarray
        2D array.
    lower_bound : float
        Lower bound.

    Returns
    -------
    x : np.ndarray
        2D array with values below lower_bound replaced by lower_bound.
    """
    if x.ndim == 1:
        x[x < lower_bound] = lower_bound
    else:
        p = x.shape[1]
        for j in range(p):
            # Numba doesn't support multidimensional index.
            x[x[:, j] < lower_bound, j] = lower_bound  # To avoid zero division
    return x


@njit
def compute_finite_difference_derivatives(ts: np.ndarray, ys: np.ndarray) -> np.ndarray:
    """
    Compute second-order finite difference derivatives.

    Without assuming uniform sampling, this function computes the second-order
    derivatives of y(t) using a finite difference approximation of the derivative.

    Parameters
    ----------
    ts : np.ndarray
        The sampling points at which to compute the derivatives. Assumed to be sorted.
    ys : np.ndarray
        The values of the function at the sampling points.

    Returns
    -------
    np.ndarray
        The approximated second-order derivatives of y(t) at the sampling points.
    """
    if len(ts) < 3:
        raise ValueError("At least three data points are required.")

    diff_weights = np.zeros((len(ts), len(ts)), dtype=np.float64)
    steps = ts[1:] - ts[:-1]

    # Second-order forward finite difference weights for the first quantile:
    first_steps_sum = steps[0] + steps[1]
    diff_weights[0, 0] = (-2 * steps[0] - steps[1]) / (steps[0] * first_steps_sum)
    diff_weights[0, 1] = first_steps_sum / (steps[0] * steps[1])
    diff_weights[0, 2] = -steps[0] / (steps[1] * first_steps_sum)

    # Central second-order finite difference weights:
    for i in range(1, len(ts) - 1):
        steps_sum = steps[i - 1] + steps[i]

        # For uniform steps, current_weight == 0.
        prev_weight = -steps[i] / (steps_sum * steps[i - 1])
        current_weight = (steps[i] - steps[i - 1]) / (steps[i] * steps[i - 1])
        next_weight = steps[i - 1] / (steps_sum * steps[i])

        diff_weights[i, i - 1] = prev_weight
        diff_weights[i, i] = current_weight
        diff_weights[i, i + 1] = next_weight

    # Second-order backward finite difference weights for the last quantile:
    last_steps_sum = steps[-2] + steps[-1]
    diff_weights[-1, -3] = (steps[-1]) / (steps[-2] * last_steps_sum)
    diff_weights[-1, -2] = (-last_steps_sum) / (steps[-2] * steps[-1])
    diff_weights[-1, -1] = (steps[-1] + last_steps_sum) / (steps[-1] * last_steps_sum)

    derivatives = np.dot(diff_weights, ys)

    return derivatives



"""Basic tests for all anomaly detectors."""

import pandas as pd
import pytest

from skchange.anomaly_detectors import (
    COLLECTIVE_ANOMALY_DETECTORS,
    StatThresholdAnomaliser,
)
from skchange.anomaly_detectors.base import BaseSegmentAnomalyDetector
from skchange.change_detectors import MovingWindow
from skchange.datasets import generate_anomalous_data

true_anomalies = [(30, 35), (70, 75)]
anomaly_data = generate_anomalous_data(
    100, anomalies=true_anomalies, means=[10.0, 15.0], random_state=2
)
anomaly_free_data = generate_anomalous_data(100, random_state=1)


@pytest.mark.parametrize("Estimator", COLLECTIVE_ANOMALY_DETECTORS)
def test_segment_anomaly_detector_predict(Estimator: BaseSegmentAnomalyDetector):
    """Test segment anomaly detector's predict method (sparse output)."""
    detector = Estimator.create_test_instance()
    detector.fit(anomaly_free_data)
    anomalies = detector.predict(anomaly_data)["ilocs"]

    assert len(anomalies) == len(true_anomalies)
    for i, (start, end) in enumerate(true_anomalies):
        assert anomalies.array.left[i] == start and anomalies.array.right[i] == end


@pytest.mark.parametrize("Estimator", COLLECTIVE_ANOMALY_DETECTORS)
def test_segment_anomaly_detector_transform(
    Estimator: BaseSegmentAnomalyDetector,
):
    """Test segment anomaly detector's transform method (dense output)."""
    detector = Estimator.create_test_instance()
    detector.fit(anomaly_free_data)
    labels = detector.transform(anomaly_data)
    true_segment_anomalies = pd.DataFrame(
        {"ilocs": pd.IntervalIndex.from_tuples(true_anomalies, closed="left")}
    )
    true_anomaly_labels = BaseSegmentAnomalyDetector.sparse_to_dense(
        true_segment_anomalies, anomaly_data.index
    )
    labels.equals(true_anomaly_labels)

    # Similar test that does not depend on sparse_to_dense, just to be sure.
    labels = labels.iloc[:, 0]
    assert labels.nunique() == len(true_anomalies) + 1
    for i, (start, end) in enumerate(true_anomalies):
        assert (labels.iloc[start:end] == i + 1).all()


def test_dense_to_sparse_invalid_columns():
    """Test dense_to_sparse method with invalid DataFrame input columns."""
    invalid_df = pd.DataFrame({"invalid_column": [0, 1, 0, 1]})
    with pytest.raises(ValueError):
        BaseSegmentAnomalyDetector.dense_to_sparse(invalid_df)


def test_stat_threshold_anomaliser_raises_stat_lower_above_stat_upper():
    """Test StatThresholdAnomaliser with valid parameters."""
    change_detector = MovingWindow(bandwidth=3)
    with pytest.raises(ValueError, match="must be less than or equal to stat_upper"):
        StatThresholdAnomaliser(
            change_detector=change_detector, stat_lower=0.5, stat_upper=0.4
        )



"""Basic tests for all change detectors."""

import pandas as pd
import pytest

from skchange.change_detectors import CHANGE_DETECTORS
from skchange.change_detectors.base import BaseChangeDetector
from skchange.datasets import generate_alternating_data

n_segments = 3
seg_len = 50
changepoint_data = generate_alternating_data(
    n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=2
)


@pytest.mark.parametrize("Estimator", CHANGE_DETECTORS)
def test_change_detector_predict(Estimator):
    """Test changepoint detector predict (sparse output)."""
    detector = Estimator.create_test_instance()
    changepoints = detector.fit_predict(changepoint_data)["ilocs"]
    assert len(changepoints) == n_segments - 1 and changepoints[0] == seg_len


@pytest.mark.parametrize("Estimator", CHANGE_DETECTORS)
def test_change_detector_transform(Estimator: BaseChangeDetector):
    """Test changepoint detector transform (dense output)."""
    detector = Estimator.create_test_instance()
    labels: pd.Series = detector.fit_transform(changepoint_data)["labels"]

    assert labels.nunique() == n_segments
    assert labels[seg_len - 1] == 0.0 and labels[seg_len] == 1.0



"""CROPS algorithm for path solutions to the PELT algorithm."""

import warnings

import numpy as np
import pandas as pd

from ..base import BaseIntervalScorer
from ..change_detectors._pelt import (
    _run_pelt,
    _run_pelt_min_segment_length_one,
    _run_pelt_with_step_size,
)
from ..change_scores._continuous_linear_trend_score import (
    lin_reg_cont_piecewise_linear_trend_score,
)
from ..costs.base import BaseCost
from ..utils.validation.interval_scorer import check_interval_scorer
from .base import BaseChangeDetector


def evaluate_segmentation(
    cost: BaseCost, segmentation: np.ndarray | pd.Series
) -> np.ndarray:
    """Evaluate the cost of a segmentation.

    Parameters
    ----------
    segmentation : np.ndarray
        A 1D array with the indices of the change points in the input data.
        Each change point signifies the first index of a new segment.

    Returns
    -------
    cost : float
        The cost of the segmentation.
    """
    cost.check_is_fitted()

    if isinstance(segmentation, pd.Series):
        segmentation = segmentation.to_numpy()
    if segmentation.ndim != 1 and segmentation.shape[1] != 1:
        raise ValueError("The segmentation must univariate")
    else:
        segmentation = segmentation.reshape(-1)

    # Prepend 0 and append the length of the data to the segmentation:
    # This is done to ensure that the first and last segments are included.
    # The segmentation is assumed to be sorted in increasing order.
    if np.any(np.diff(segmentation) <= 0):
        raise ValueError("The segmentation must contain strictly increasing entries.")
    if len(segmentation) == 0:
        segmentation = np.array([0, cost.n_samples])
    elif segmentation[0] != 0 and segmentation[-1] != cost.n_samples:
        segmentation = np.concatenate(
            (np.array([0]), segmentation, np.array([cost.n_samples]))
        )

    cuts = np.vstack((segmentation[:-1], segmentation[1:])).T

    # Aggregate the partition intervals into a 1D array of values:
    return np.sum(cost.evaluate(cuts), axis=0)


def _format_crops_results(
    penalty_to_solution_dict: dict[float, tuple[np.ndarray, float]],
    penalty_nudge: float,
) -> tuple[pd.DataFrame, dict]:
    """Format the CROPS results into a DataFrame.

    Parameters
    ----------
    cpd_results : dict
        Dictionary with penalty values as keys and tuples of change points and
        segmentation costs as values.

    Returns
    -------
    tuple[pd.DataFrame, dict]
        DataFrame with columns 'num_change_points', 'penalty', and 'segmentation_cost'.
        Dictionary with number of change points as keys and change points as values.
    """
    # Convert the dict lookup results to a list of tuples:
    list_cpd_results = [
        (len(change_points), penalty, segmentation_cost, change_points)
        for (
            penalty,
            (change_points, segmentation_cost),
        ) in penalty_to_solution_dict.items()
    ]

    # When different penalties give the same number of change points, we want to
    # keep the segmentation with the lowest penalty. So we sort by number of
    # change points, and then negative penalty, in reverse order.
    list_cpd_results.sort(key=lambda x: (x[0], -x[1]), reverse=True)

    # If there are solutions that only differ by a single changepoint when we increase
    # the penalty, we can compute the penalty for which the number of change points
    # changes when increasing penalty from the lower value to the higher value.
    for i in range(len(list_cpd_results) - 1):
        # If the number of change points is the same, and the penalty is different,
        # we can compute the penalty for which the number of change points changes
        # when increasing penalty from the lower value to the higher value.
        lower_penalty_num_change_points = list_cpd_results[i][0]
        higher_penalty_num_change_points = list_cpd_results[i + 1][0]
        if lower_penalty_num_change_points == higher_penalty_num_change_points + 1:
            # Difference in segmentation cost is the critical penalty value.
            # Multiply by (1.0 + penalty_nudge) to avoid numerical instability.
            lower_penalty_segmentation_cost = list_cpd_results[i][2]
            higher_penalty_segmentation_cost = list_cpd_results[i + 1][2]
            critical_penalty = (
                higher_penalty_segmentation_cost - lower_penalty_segmentation_cost
            ) * (1.0 + penalty_nudge)

            # Replace the penalty value with the critical penalty value.
            list_cpd_results[i + 1] = (
                list_cpd_results[i + 1][0],
                critical_penalty,
                list_cpd_results[i + 1][2],
                list_cpd_results[i + 1][3],
            )

    # Remove duplicates, and keep the first one:
    encountered_num_change_points = set()
    unique_cpd_results = []
    for i in range(len(list_cpd_results)):
        num_change_points = list_cpd_results[i][0]
        if num_change_points in encountered_num_change_points:
            continue
        else:
            unique_cpd_results.append(list_cpd_results[i])
            encountered_num_change_points.add(num_change_points)

    # Extract out the change points into a dict with the number of change points
    # as the key, and the change points as the value.
    change_points_dict = {len(x[3]): x[3] for x in unique_cpd_results}

    # Extract out the change points metadata into a DataFrame.
    # Reverse the order of the list, so that the lowest number
    # of change points comes first.
    change_points_metadata = [x[0:3] for x in unique_cpd_results][::-1]
    penalty_change_point_metadata_df = pd.DataFrame(
        change_points_metadata,
        columns=[
            "num_change_points",
            "penalty",
            "segmentation_cost",
        ],
    )

    # Ensure that the number of change points is an integer:
    penalty_change_point_metadata_df["num_change_points"] = (
        penalty_change_point_metadata_df["num_change_points"].astype(int)
    )

    penalty_change_point_metadata_df["optimum_value"] = (
        penalty_change_point_metadata_df["segmentation_cost"]
        + penalty_change_point_metadata_df["penalty"]
        * penalty_change_point_metadata_df["num_change_points"]
    )

    return penalty_change_point_metadata_df, change_points_dict


def segmentation_bic_value(cost: BaseCost, change_points: np.ndarray) -> float:
    """Calculate the BIC score for a given segmentation.

    Parameters
    ----------
    cost : BaseCost
        The cost function to use.
    change_points : np.ndarray
        The change points to use.

    Returns
    -------
    float
        The BIC score for the given segmentation.
    """
    cost.check_is_fitted()
    num_segments = len(change_points) + 1
    data_dim = cost.n_variables
    num_parameters = cost.get_model_size(data_dim)
    n_samples = cost.n_samples

    bic_score = evaluate_segmentation(cost, change_points)[
        0
    ] + num_parameters * num_segments * np.log(n_samples)

    return bic_score


def crops_elbow_scores(
    num_change_points_and_optimum_value_df: pd.DataFrame,
) -> np.ndarray:
    """Calculate the elbow cost for a given segmentation.

    Specifically, the elbow score is calculated as the evidence for a change
    of slope in the segmentation cost as a function of the number of
    change points, at each intermediate number of change points.
    We cannot calculate the elbow score for the first and last number of
    change points, as there are not enough segmentations to calculate a change
    in slope before or after the first and last number of change points.

    Parameters
    ----------
    num_change_points : int
        The number of change points.
    segmentation_cost : float
        The segmentation cost.

    Returns
    -------
    pd.Series
        The elbow cost for each number of change points.
    """
    num_segmentations = len(num_change_points_and_optimum_value_df)
    if num_segmentations < 3:
        # Not enough segmentations to calculate the elbow cost.
        warnings.warn(
            f"Not enough segmentations {num_segmentations} to calculate "
            "the elbow cost. Returning -np.inf for all segmentations."
        )
        return pd.Series([-np.inf] * num_segmentations)

    # Calculate the elbow (change in slope) cost for each number of change points:
    times = num_change_points_and_optimum_value_df["num_change_points"].values
    optim_values = num_change_points_and_optimum_value_df[
        "optimum_value"
    ].values.reshape(-1, 1)
    elbow_values = lin_reg_cont_piecewise_linear_trend_score(
        starts=np.repeat(0, num_segmentations - 2),
        splits=np.arange(1, num_segmentations - 1),
        ends=np.repeat(num_segmentations, num_segmentations - 2),
        X=optim_values,
        times=times,
    )

    # Pad the elbow score with `-np.inf` for the first and last segmentations,
    # which we cannot calculate a "change in slope" score for.
    elbow_scores = np.concatenate(
        (
            np.array([-np.inf]),
            elbow_values.reshape(-1),
            np.array([-np.inf]),
        )
    )

    return elbow_scores


class CROPS(BaseChangeDetector):
    """CROPS algorithm for path solutions to the `PELT` algorithm.

    This change detector solves for all penalized optimal partitionings
    within the penalty range `[min_penalty, max_penalty]`, using the CROPS
    algorithm[1]_, which in turn employs the `PELT` algorithm to repeatedly
    solve penalized optimal partitioning problems for different penalties.

    When predicting change points through `predict()`, this change detector
    selects the best segmentation among the optimal partitionings within
    the penalty range, using the `selection_method` criterion.

    Parameters
    ----------
    cost : BaseIntervalScorer
        The cost function to use.
    min_penalty : float
        The start of the penalty solution interval.
    max_penalty : float
        The end of the penalty solution interval.
    selection_method : str, default="bic"
        The selection method to use when selecting the
        best segmentation among the optimal segmentations found
        within the penalty range `[min_penalty, max_penalty]`.
        The options are:

        * ``"bic"``: Select the segmentation with the lowest Bayesian Information
          criterion, defined as `segmentation_cost + model_size_per_segment *
          log(n_samples)`.
        * ``"elbow"``: Select the segmentation with the highest elbow score.
          Defined as the improvement in squared residuals when allowing
          a change in slope of segmentation cost regressed on the number
          of change points, at each intermediate number of change points.

    min_segment_length : int, default=1
        The minimum segment length to use.
    split_cost : float, optional, default=0.0
        The cost of splitting a segment, to ensure that
        cost(X[t:p]) + cost(X[p:(s+1)]) + split_cost <= cost(X[t:(s+1)]),
        for all possible splits, 0 <= t < p < s <= len(X) - 1.
        By default set to 0.0, which is sufficient for
        log likelihood cost functions to satisfy the above inequality.
    prune: bool, optional
        If False, drop the pruning step, reverting to optimal partitioning.
        Can be useful for debugging and testing. By default set to True.
    pruning_margin : float, optional, default=0.0
        The pruning margin to use. By default set to zero.
        This is used to reduce pruning of the admissible starts set.
        Can be useful when the cost function is imprecise, e.g.
        based on solving an optimization problem with large tolerance.
    middle_penalty_nudge : float, optional, default=1.0e-4
        When computing the threshold penalty value separating `PELT `solutions
        with differing numbers of change points, we need to nudge the penalty
        upwards in order to solve for the segmentation with fewer change points.
        By default set to 1.0e-4, which is sufficient for most cases.

    References
    ----------
    .. [1] Haynes, K., Eckley, I. A., & Fearnhead, P. (2017). Computationally efficient
    changepoint detection for a range of penalties. Journal of Computational and
    Graphical Statistics, 26(1), 134-143.
    """

    _tags = {
        "authors": ["johannvk"],
        "maintainers": ["johannvk"],
        "fit_is_empty": True,
    }

    def __init__(
        self,
        cost: BaseIntervalScorer,
        min_penalty: float,
        max_penalty: float,
        segmentation_selection: str = "bic",
        min_segment_length: int = 1,
        step_size: int = 1,
        split_cost: float = 0.0,
        prune: bool = True,
        pruning_margin: float = 0.0,
        middle_penalty_nudge: float = 1.0e-5,
    ):
        super().__init__()
        self.cost = cost
        self.min_penalty = min_penalty
        self.max_penalty = max_penalty
        self.segmentation_selection = segmentation_selection
        self.min_segment_length = min_segment_length
        self.step_size = step_size
        self.split_cost = split_cost
        self.pruning_margin = pruning_margin
        self.middle_penalty_nudge = middle_penalty_nudge
        self.prune = prune

        self._cost = cost.clone()
        check_interval_scorer(
            self._cost,
            arg_name="cost",
            caller_name="CROPS",
            required_tasks=["cost"],
            allow_penalised=False,
        )

        if self.step_size > 1 and self.min_segment_length > self.step_size:
            raise ValueError(
                f"CROPS `min_segment_length`(={self.min_segment_length}) cannot be "
                f"greater than the `step_size`(={self.step_size}) > 1."
            )

        if segmentation_selection not in ["bic", "elbow"]:
            raise ValueError(
                f"Invalid selection criterion: {segmentation_selection}. "
                "Must be one of ['bic', 'elbow']."
            )

        # Storage for the CROPS results:
        self.change_points_metadata: pd.DataFrame | None = None
        self.change_points_lookup: dict[int, np.ndarray] | None = None
        self.optimal_penalty: float | None = None

    def _predict(self, X: np.ndarray) -> np.ndarray:
        """Run the CROPS algorithm for path solutions to penalized CPD.

        Parameters
        ----------
        X : np.ndarray
            Data to search for change points in.

        Returns
        -------
        np.ndarray
            The optimal change points for penalties within
            `[self.min_penalty, self.max_penalty]`, as decided by the
            `segmentation_selection` criterion.

        Attributes
        ----------
        optimal_penalty : float
            The penalty value for which the optimal change points were found,
            w.r.t. the `segmentation_selection` criterion.
        """
        if X.ndim > 1 and X.shape[1] > 1 and not self._cost.get_tag("is_aggregated"):
            raise ValueError(
                "CROPS only supports costs that return a single value per cut "
                "when the input data has more than one column. "
                "Please use an aggregated cost function."
            )
        self._run_crops(X=X)

        if self.segmentation_selection == "elbow":
            # Select the best segmentation using the elbow criterion.
            change_in_slope_df = self.change_points_metadata[
                ["num_change_points", "optimum_value"]
            ].copy()

            # Subrtract the minimum value from the optimum value
            # to improve conditioning number of linreg problems.
            change_in_slope_df["optimum_value"] = (
                change_in_slope_df["optimum_value"]
                - change_in_slope_df["optimum_value"].min()
            )
            self.change_points_metadata["elbow_score"] = crops_elbow_scores(
                change_in_slope_df,
            )
            optimal_num_change_points, optimal_penalty = (
                self.change_points_metadata.sort_values(
                    by="elbow_score", ascending=False
                )[["num_change_points", "penalty"]].iloc[0]
            )

        elif self.segmentation_selection == "bic":
            # Select the best segmentation using the BIC criterion.
            self.change_points_metadata["bic_value"] = self.change_points_metadata[
                "num_change_points"
            ].apply(
                lambda num_change_points: segmentation_bic_value(
                    cost=self._cost,
                    change_points=self.change_points_lookup[num_change_points],
                )
            )
            optimal_num_change_points, optimal_penalty = (
                self.change_points_metadata.sort_values(
                    by="bic_value"
                )[["num_change_points", "penalty"]].iloc[0]
            )

        self.optimal_penalty = optimal_penalty
        return self.change_points_lookup[optimal_num_change_points]

    def _solve_for_changepoints(self, penalty: float) -> np.ndarray:
        """Solve for the optimal changepoints given `penalty` using PELT.

        Parameters
        ----------
        X : np.ndarray
            Data to search for change points in.

        Returns
        -------
        changepoints : np.ndarray
            The optimal change points for the given penalty.
        """
        if self.step_size > 1:
            pelt_result = _run_pelt_with_step_size(
                self._cost,
                penalty=penalty,
                step_size=self.step_size,
                split_cost=self.split_cost,
                prune=self.prune,
                pruning_margin=self.pruning_margin,
            )
        elif self.min_segment_length == 1:
            pelt_result = _run_pelt_min_segment_length_one(
                self._cost,
                penalty=penalty,
                split_cost=self.split_cost,
                prune=self.prune,
                pruning_margin=self.pruning_margin,
            )
        else:
            pelt_result = _run_pelt(
                self._cost,
                penalty=penalty,
                min_segment_length=self.min_segment_length,
                split_cost=self.split_cost,
                prune=self.prune,
                pruning_margin=self.pruning_margin,
            )

        return pelt_result.changepoints

    def _run_crops(self, X: np.ndarray) -> pd.DataFrame:
        """Run the CROPS algorithm for path solutions to penalized CPD on data `X`.

        Parameters
        ----------
        X : np.ndarray
            Data to search for change points in.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns 'num_change_points', 'penalty',
            and 'segmentation_cost'.

        Attributes
        ----------
        change_points_metadata : pd.DataFrame
            DataFrame with columns 'num_change_points', 'penalty',
            and 'segmentation_cost'.
            Contains metadata about the change points found by the CROPS algorithm.
        change_points_lookup : dict[int, np.ndarray]
            Dictionary with number of change points as keys and change points as values.
            The keys are the number of change points, and the values are arrays with the
            change point indices for that number of change points.
        """
        self._cost.fit(X)

        min_penalty_change_points = self._solve_for_changepoints(
            penalty=self.min_penalty
        )
        max_penalty_change_points = self._solve_for_changepoints(
            penalty=self.max_penalty
        )

        num_min_penalty_change_points = len(min_penalty_change_points)
        num_max_penalty_change_points = len(max_penalty_change_points)

        min_penalty_segmentation_cost = evaluate_segmentation(
            self._cost, min_penalty_change_points
        )[0]
        max_penalty_segmentation_cost = evaluate_segmentation(
            self._cost, max_penalty_change_points
        )[0]

        penalty_to_solution_dict: dict[float, (np.ndarray, float)] = dict()
        penalty_to_solution_dict[self.min_penalty] = (
            min_penalty_change_points,
            min_penalty_segmentation_cost,
        )
        penalty_to_solution_dict[self.max_penalty] = (
            max_penalty_change_points,
            max_penalty_segmentation_cost,
        )

        # Store intervals of penalty values in which to search for
        # differing numbers of change points.
        penalty_search_intervals = []
        if num_min_penalty_change_points > num_max_penalty_change_points + 1:
            # More than one change point in difference between min and max penalty.
            # Need to split the penalty intervals further.
            penalty_search_intervals.append((self.min_penalty, self.max_penalty))

        while len(penalty_search_intervals) > 0:
            # Pop the interval with the lowest penalty.
            low_penalty, high_penalty = penalty_search_intervals.pop(0)

            low_penalty_change_points, low_penalty_segmentation_cost = (
                penalty_to_solution_dict[low_penalty]
            )
            num_low_penalty_change_points = len(low_penalty_change_points)

            high_penalty_change_points, high_penalty_segmentation_cost = (
                penalty_to_solution_dict[high_penalty]
            )
            num_high_penalty_change_points = len(high_penalty_change_points)

            if num_low_penalty_change_points > (num_high_penalty_change_points + 1):
                # Compute the threshold penalty value, where the number of change
                # points decreases.
                threshold_penalty = (
                    high_penalty_segmentation_cost - low_penalty_segmentation_cost
                ) / (num_low_penalty_change_points - num_high_penalty_change_points)

                # Nudge the middle penalty towards the high penalty, to ensure that
                # the number of change points for the middle penalty is fewer than
                # for the low penalty.
                additive_nudged_middle_penalty = (
                    threshold_penalty
                    + (high_penalty - threshold_penalty) * self.middle_penalty_nudge
                )
                multiplicative_nudged_middle_penalty = threshold_penalty * (
                    1.0 + self.middle_penalty_nudge
                )
                # Middle penalty is the minimum of the two nudged values:
                middle_penalty = min(
                    additive_nudged_middle_penalty, multiplicative_nudged_middle_penalty
                )

                middle_penalty_change_points = self._solve_for_changepoints(
                    penalty=middle_penalty
                )
                middle_penalty_segmentation_cost = evaluate_segmentation(
                    self._cost, middle_penalty_change_points
                )[0]
                penalty_to_solution_dict[middle_penalty] = (
                    middle_penalty_change_points,
                    middle_penalty_segmentation_cost,
                )

                middle_penalty_matches_high_penalty = len(
                    middle_penalty_change_points
                ) == len(high_penalty_change_points)
                middle_penalty_matches_low_penalty = len(
                    middle_penalty_change_points
                ) == len(low_penalty_change_points)

                if middle_penalty_matches_high_penalty:
                    # The same number of change points for penalties in
                    # the interval [middle_penalty, high_penalty].
                    # Don't need to subdivide penalty intervals further.
                    continue
                elif middle_penalty_matches_low_penalty:
                    raise ValueError(  # pragma: no cover
                        "PELT optimization has not been solved exactly! "
                        "Number of change points should be greater for the "
                        "middle penalty than for the low penalty. "
                        "Attempt to set the `split_cost` parameter to a "
                        "non-zero value, or increase `pruning_margin`."
                    )
                else:
                    # Number of change points for middle penalty is different from both
                    # low_penalty and high_penalty. Need to explore further.
                    penalty_search_intervals.append((low_penalty, middle_penalty))
                    penalty_search_intervals.append((middle_penalty, high_penalty))

                    # Sort the intervals by lower penalty:
                    penalty_search_intervals.sort(key=lambda x: x[0])

        metadata_df, change_points_dict = _format_crops_results(
            penalty_to_solution_dict=penalty_to_solution_dict,
            penalty_nudge=self.middle_penalty_nudge,
        )
        self.change_points_metadata = metadata_df
        self.change_points_lookup = change_points_dict

        return self.change_points_metadata

    @classmethod
    def get_test_params(cls, parameter_set: str = "default") -> list[dict]:
        """Get test parameters for the CROPS algorithm."""
        from skchange.costs import L2Cost

        return [
            {
                "cost": L2Cost(),
                "min_penalty": 0.5,
                "max_penalty": 50.0,
                "segmentation_selection": "bic",
                "min_segment_length": 10,
                "split_cost": 0.0,
                "pruning_margin": 0.0,
                "middle_penalty_nudge": 1.0e-4,
                "prune": True,
            },
            {
                "cost": L2Cost(),
                "min_penalty": 1.0,
                "max_penalty": 10.0,
                "segmentation_selection": "elbow",
                "min_segment_length": 2,
                "split_cost": 0.0,
                "pruning_margin": 0.01,
                "middle_penalty_nudge": 1.0e-4,
                "prune": False,
            },
        ]



import numpy as np
import pytest

from skchange.costs import GaussianCost


@pytest.mark.parametrize(
    "mean, var",
    [
        (0.0, np.array([1.0, 2.0, 3.0])),  # Too many entries compared to data
        (0.0, -1.0),  # Negative variance
    ],
)
def test_invalid_fixed_covariance(mean, var):
    """Test that invalid fixed covariance matrix raises errors."""
    X = np.random.randn(100, 2)
    cost = GaussianCost(param=(mean, var))
    with pytest.raises(ValueError):
        cost.fit(X)



import numpy as np
import pytest
from skbase._exceptions import NotFittedError

from skchange.costs import COSTS
from skchange.costs.base import BaseCost
from skchange.datasets import generate_alternating_data
from skchange.tests.test_all_interval_scorers import skip_if_no_test_data


def find_fixed_param_combination(
    cost_class: type[BaseCost] | BaseCost,
) -> dict[str, float]:
    """Find the first fixed parameter combination in the test parameters of a cost."""
    test_param_sets = cost_class.get_test_params()
    fixed_test_param_set = None
    for param_set in test_param_sets:
        if "param" in param_set and param_set["param"] is not None:
            fixed_test_param_set = param_set
            break

    if fixed_test_param_set is None:
        class_name = (
            cost_class.__name__
            if isinstance(cost_class, type)
            else cost_class.__class__.__name__
        )
        raise ValueError(
            f"No fixed `param` argument found in `get_test_params()` of"
            f" the cost class {class_name}"
        )

    return fixed_test_param_set


def create_fixed_cost_test_instance(cost_class: type[BaseCost]) -> BaseCost:
    """Create a fixed instance of the cost class."""
    fixed_param = find_fixed_param_combination(cost_class)
    return cost_class.create_test_instance().set_params(**fixed_param)


def test_find_fixed_param_combination_value_error():
    class MockCost:
        @staticmethod
        def get_test_params():
            return [{"param": None}, {"param": None}]

    with pytest.raises(ValueError):
        find_fixed_param_combination(MockCost)


@pytest.mark.parametrize("CostClass", COSTS)
def test_l2_cost_init(CostClass: type[BaseCost]):
    cost = CostClass.create_test_instance()
    assert cost.param is None


@pytest.mark.parametrize("CostClass", COSTS)
def test_expected_cut_entries(CostClass: type[BaseCost]):
    cost = CostClass.create_test_instance()
    assert cost._get_required_cut_size() == 2


@pytest.mark.parametrize("CostClass", COSTS)
def test_cost_evaluation_optim_gt_fixed(CostClass: type[BaseCost]):
    if not CostClass.get_class_tag("supports_fixed_param"):
        pytest.skip(f"{CostClass.__name__} does not support fixed parameters.")

    optim_cost = CostClass.create_test_instance()
    skip_if_no_test_data(optim_cost)
    fixed_cost = create_fixed_cost_test_instance(CostClass)
    np.random.seed(1001)
    X = np.random.multivariate_normal(mean=[0, 0], cov=[[1, 0], [0, 1]], size=40)
    optim_cost.fit(X)
    fixed_cost.fit(X)
    intervals = np.array([[0, 15], [10, 30], [5, 25], [25, 40]])
    optim_costs = optim_cost.evaluate(intervals)
    fixed_costs = fixed_cost.evaluate(intervals)
    assert np.all(optim_costs <= fixed_costs)


@pytest.mark.parametrize("CostClass", COSTS)
def test_cost_evaluation_positive(CostClass: type[BaseCost]):
    cost = CostClass.create_test_instance()
    skip_if_no_test_data(cost)
    n = 50
    df = generate_alternating_data(n_segments=1, segment_length=n, p=1, random_state=5)
    cost.fit(df)
    starts = np.arange(n - 15)
    ends = np.repeat(n - 1, len(starts))
    intervals = np.column_stack((starts, ends))
    costs = cost.evaluate(intervals)
    assert np.all(costs >= 0.0)


@pytest.mark.parametrize("CostClass", COSTS)
def test_accessing_n_samples_before_fit_raises(
    CostClass: type[BaseCost],
):
    """Test that accessing n_samples before fitting raises an error."""
    cost = CostClass.create_test_instance()
    with pytest.raises(NotFittedError):
        cost.n_samples


@pytest.mark.parametrize("CostClass", COSTS)
def test_accessing_n_variables_before_fit_raises(
    CostClass: type[BaseCost],
):
    """Test that accessing n_variables before fitting raises an error."""
    cost = CostClass.create_test_instance()
    with pytest.raises(NotFittedError):
        cost.n_variables



"""Tests for CircularBinarySegmentation."""

import pytest

from skchange.anomaly_detectors import CircularBinarySegmentation
from skchange.change_scores import ChangeScore
from skchange.costs import COSTS


def test_invalid_change_scores():
    """
    Test that CircularBinarySegmentation raises an error when given an invalid score.
    """
    with pytest.raises(ValueError, match="anomaly_score"):
        CircularBinarySegmentation("l2")
    with pytest.raises(ValueError, match="anomaly_score"):
        CircularBinarySegmentation(ChangeScore(COSTS[2].create_test_instance()))



"""Extension template for costs as interval scorers.

Adapted from the sktime extension templates.

Purpose of this implementation template:
    quick implementation of new estimators following the template
    NOT a concrete class to import! This is NOT a base class or concrete class!
    This is to be used as a "fill-in" coding template.

How to use this implementation template to implement a new estimator:
- make a copy of the template in a suitable location, give it a descriptive name.
- work through all the "todo" comments below
- fill in code for mandatory methods, and optionally for optional methods
- you can add more private methods, but do not override BaseCost's private methods
    an easy way to be safe is to prefix your methods with "_custom"
- change docstrings for functions and the file
- ensure interface compatibility by sktime.utils.estimator_checks.check_estimator
- once complete: use as a local library, or contribute to skchange via PR
- more details:
  https://www.sktime.net/en/stable/developer_guide/add_estimators.html

Mandatory implements:
    fitting                  - _fit(self, X, y=None)
    evaluating optimal param - _evaluate_optim_param(self, starts, ends)

Optional implements:
    evaluating fixed param   - _evaluate_fixed_param(self, starts, ends)
    checking fixed param     - _check_fixed_param(self, param, X)
    minimum size of interval  - min_size(self)
    number of parameters      - get_model_size(self, p)

Testing - required for sktime test framework and check_estimator usage:
    get default parameters for test instance(s) - get_test_params()

copyright: skchange developers, BSD-3-Clause License (see LICENSE file)
"""

# todo: add any necessary imports

import numpy as np

# internal extensions located in skchange.costs:
from .base import BaseCost

# external extensions:
# from skchange.costs.base import BaseCost

# todo: add the cost to the COSTS variable in skchange.costs.__init__.py


class MyCost(BaseCost):
    """Custom cost class.

    todo: write docstring, describing your custom cost.

    Parameters
    ----------
    param : any, optional (default=None)
        If None, the cost is evaluated for an interval-optimised parameter, often the
        maximum likelihood estimate. If not None, the cost is evaluated for the
        specified fixed parameter.
    param1 : string, optional, default=None
        descriptive explanation of param1
    param2 : float, optional, default=1.0
        descriptive explanation of param2
    and so on
    """

    # todo: add authors and maintaners Github user name
    _tags = {
        "authors": ["Tveten", "johannvk"],  # Use Github username
        "maintainers": "Tveten",  # Use Github username
        # distribution_type is used to automatically create test cases.
        # Valid values:
        # - "None" - No distributional restrictions. Test data: Mostly Gaussian, but
        #   no guarantee.
        # - "Poisson" - Integer data. Test data: Poisson distributed.
        # - "Gaussian" - Real-valued data. Test data: Gaussian distribution.
        "distribution_type": "None",  # "None", "Poisson", "Gaussian"
        # is_conditional: whether the scorer uses some of the input variables as
        # covariates in a regression model or similar. If `True`, the scorer requires
        # at least two input variables. If `False`, all p input variables/columns are
        # used to evaluate the score, such that the output has either 1 or p columns.
        "is_conditional": False,
        # is_aggregated: whether the scorer always returns a single value per cut or
        # not, irrespective of the input data shape.
        # Many scorers will not be aggregated, for example all scorers that evaluate
        # each input variable separately and return a score vector with one score for
        # each variable.
        "is_aggregated": False,
        # Does the cost support fixed parameters? I.e., is `_evaluate_fixed_param`
        # implemented? Fixed parameter evaluation is required for the cost to be used as
        # a saving in certain anomaly detectors.
        "supports_fixed_param": False,
    }

    # todo: add any hyper-parameters and components to constructor
    # todo: if fixed parameters are supported, add type hints to `param`
    def __init__(
        self,
        param=None,  # Mandatory first parameter (see docs above).
        param1=None,  # Custom parameter 1.
        param2=1.0,  # Custom parameter 2.
    ):
        # leave this as is.
        super().__init__(param)

        # todo: write any hyper-parameters and components to self. These should never
        # be overwritten in other methods.
        # estimators should precede parameters
        #  if estimators have default values, set None and initialize below
        self.param1 = param1
        self.param2 = param2

        # todo: optional, parameter checking logic (if applicable) should happen here
        # if writes derived values to self, should *not* overwrite self.parama etc
        # instead, write to self._param1, etc.
        if self.param2 is None:
            from skchange.somewhere import MyOtherCost

            self._param2 = MyOtherCost(foo=42)
        else:
            # estimators should be cloned to avoid side effects
            self._param2 = param2.clone()

        # todo: if tags of estimator depend on component tags, set these here
        #  only needed if estimator is a composite
        #  tags set in the constructor apply to the object and override the class
        #
        # example 1: conditional setting of a tag
        # if est.foo == 42:
        #   self.set_tags(handles-missing-data=True)
        # example 2: cloning tags from component
        #   self.clone_tags(est2, ["enforce_index_type", "handles-missing-data"])

    # todo: implement, mandatory
    def _fit(self, X: np.ndarray, y=None):
        """Fit the cost.

        This method precomputes quantities that speed up the cost evaluation.

        Parameters
        ----------
        X : np.ndarray
            Data to evaluate. Must be a 2D array.
        y: None
            Ignored. Included for API consistency by convention.
        """
        self._param = self._check_param(self.param, X)

        # todo: implement precomputations here
        # IMPORTANT: avoid side effects to X, y.

        # for example for the L2 cost:
        # self._sums = col_cumsum(X, init_zero=True)
        # self._sums2 = col_cumsum(X**2, init_zero=True)

        return self

    # todo: implement, mandatory
    def _evaluate_optim_param(self, starts: np.ndarray, ends: np.ndarray) -> np.ndarray:
        """Evaluate the cost for the optimal parameters.

        Evaluates the cost for `X[start:end]` for each each start, end in starts, ends.

        Parameters
        ----------
        starts : np.ndarray
            Start indices of the intervals (inclusive).
        ends : np.ndarray
            End indices of the intervals (exclusive).

        Returns
        -------
        costs : np.ndarray
            A 2D array of costs. One row for each interval. The number of
            columns is 1 if the cost has `evaluation_type = "multivariate"`.
            The number of columns is equal to the number of columns in the input data if
            the `evaluation_type = "univariate"`. In this case, each column represents
            the univariate cost for the corresponding input data column.
        """
        # todo: implement evaluation logic here. Must have the output format as
        #       described in the docstring.
        # IMPORTANT: avoid side effects to starts, ends.

    # todo: implement, optional. Mandatory if `supports_fixed_param` = True.
    def _evaluate_fixed_param(self, starts, ends) -> np.ndarray:
        """Evaluate the cost for the fixed parameters.

        Evaluates the cost for `X[start:end]` for each each start, end in starts, ends.

        Parameters
        ----------
        starts : np.ndarray
            Start indices of the intervals (inclusive).
        ends : np.ndarray
            End indices of the intervals (exclusive).

        Returns
        -------
        costs : np.ndarray
            A 2D array of costs. One row for each interval. The number of
            columns is 1 if the cost has `evaluation_type = "multivariate"`.
            The number of columns is equal to the number of columns in the input data if
            the `evaluation_type = "univariate"`. In this case, each column represents
            the univariate cost for the corresponding input data column.
        """
        # todo: implement evaluation logic here. Must have the output format as
        #       described in the docstring.
        # IMPORTANT: avoid side effects to starts, ends.

    # todo: implement, optional, defaults to no checking.
    # used inside self._check_param in _fit.
    def _check_fixed_param(self, param, X: np.ndarray) -> np.ndarray:
        """Check if the fixed parameter is valid relative to the data.

        Parameters
        ----------
        param : any
            Fixed parameter for the cost calculation.
        X : np.ndarray
            Input data.

        Returns
        -------
        param: any
            Fixed parameter for the cost calculation.
        """
        return param

    # todo: implement, optional, defaults to min_size = 1.
    # used for automatic validation of cuts in `evaluate`.
    @property
    def min_size(self) -> int | None:
        """Minimum size of the interval to evaluate.

        The size of each interval is defined as ``cuts[i, 1] - cuts[i, 0]``.

        Returns
        -------
        int or None
            The minimum valid size of an interval to evaluate. If ``None``, it is
            unknown what the minimum size is. E.g., the scorer may need to be fitted
            first to determine the minimum size.
        """
        # For example for a mean and variance cost:
        # return 2
        #
        # For example for a covariance matrix cost:
        # if self.is_fitted:
        #     return self.n_variables + 1
        # else:
        #     return None

    # todo: implement, optional, defaults to output p (one parameter per variable).
    # used for setting a decent default penalty in detectors.
    def get_model_size(self, p: int) -> int:
        """Get the number of parameters in the cost function.

        Parameters
        ----------
        p : int
            Number of variables in the data.

        Returns
        -------
        int
            Number of parameters in the cost function.
        """
        # For example for a covariance matrix cost:
        # return p * (p + 1) // 2

    # todo: return default parameters, so that a test instance can be created.
    #       required for automated unit and integration testing of estimator.
    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for costs.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """

        # todo: set the testing parameters for the estimators
        # Testing parameters can be dictionary or list of dictionaries
        # Testing parameter choice should cover internal cases well.
        #
        # this method can, if required, use:
        #   class properties (e.g., inherited); parent class test case
        #   imported objects such as estimators from sktime or sklearn
        # important: all such imports should be *inside get_test_params*, not at the top
        #            since imports are used only at testing time
        #
        # The parameter_set argument is not used for automated, module level tests.
        #   It can be used in custom, estimator specific tests, for "special" settings.
        # A parameter dictionary must be returned *for all values* of parameter_set,
        #   i.e., "parameter_set not available" errors should never be raised.
        #
        # A good parameter set should primarily satisfy two criteria,
        #   1. Chosen set of parameters should have a low testing time,
        #      ideally in the magnitude of few seconds for the entire test suite.
        #       This is vital for the cases where default values result in
        #       "big" models which not only increases test time but also
        #       run into the risk of test workers crashing.
        #   2. There should be a minimum two such parameter sets with different
        #      sets of values to ensure a wide range of code coverage is provided.
        #
        # example 1: specify params as dictionary
        # any number of params can be specified
        # params = {"est": value0, "parama": value1, "paramb": value2}
        # return params
        #
        # example 2: specify params as list of dictionary
        # note: Only first dictionary will be used by create_test_instance
        # params = [{"est": value1, "parama": value2},
        #           {"est": value3, "parama": value4}]
        # return params
        #
        # example 3: parameter set depending on param_set value
        #   note: only needed if a separate parameter set is needed in tests
        # if parameter_set == "special_param_set":
        #     params = {"est": value1, "parama": value2}
        #     return params
        #
        # # "default" params - always returned except for "special_param_set" value
        # params = {"est": value3, "parama": value4}
        # return params



"""Tests for the PELT implementation."""

import re
import time

import numpy as np
import pytest
import ruptures as rpt
from ruptures.base import BaseCost as rpt_BaseCost

from skchange.change_detectors._crops import evaluate_segmentation
from skchange.change_detectors._pelt import (
    PELT,
    PELTResult,
    _run_pelt,
    _run_pelt_min_segment_length_one,
    _run_pelt_with_step_size,
    get_changepoints,
)
from skchange.change_scores import CUSUM
from skchange.costs import GaussianCost, L2Cost
from skchange.costs.base import BaseCost
from skchange.datasets import generate_alternating_data

n_segments = 2
seg_len = 50
changepoint_data = generate_alternating_data(
    n_segments=n_segments, mean=20, segment_length=seg_len, p=1, random_state=2
).values.reshape(-1, 1)

alternating_sequence = generate_alternating_data(
    n_segments=5,
    mean=10.5,
    variance=0.5,
    segment_length=21,
    p=1,
    random_state=5,
).values.reshape(-1, 1)

long_alternating_sequence = generate_alternating_data(
    n_segments=30,
    mean=10.5,
    variance=0.5,
    segment_length=21,
    p=1,
    random_state=5,
).values.reshape(-1, 1)


@pytest.fixture
def cost():
    """Generate a new cost object for each test."""
    cost = L2Cost()
    return cost


@pytest.fixture
def penalty() -> float:
    """Penalty for the PELT algorithm."""
    penalty = 2 * np.log(len(changepoint_data))
    return penalty


class RupturesGaussianCost(rpt_BaseCost):
    """Custom cost for Gaussian (mean-var) cost."""

    # The 2 following attributes must be specified for compatibility.
    model = ""
    min_size = 2

    def fit(self, signal) -> "RupturesGaussianCost":
        self.signal = signal
        self.cost = GaussianCost().fit(signal)
        return self

    def error(self, start: int, end: int) -> np.ndarray:
        """Compute the cost of a segment."""
        cuts = np.array([start, end]).reshape(-1, 2)
        return self.cost.evaluate(cuts)


def pelt_partition_cost(
    X: np.ndarray,
    changepoints: np.ndarray,
    cost: BaseCost,
    penalty: float,
):
    cost.fit(X)
    num_samples = len(X)

    # Add number of 'segments' * penalty to the cost.
    # Instead of number of 'changepoints' * penalty.
    # total_cost = penalty * (len(changepoints) + 1)
    total_cost = penalty * len(changepoints)
    np_changepoints = np.asarray(changepoints, dtype=np.int64)

    interval_starts = np.concatenate((np.array([0]), np_changepoints), axis=0)
    interval_ends = np.concatenate((np_changepoints, np.array([num_samples])), axis=0)

    interval_costs = np.sum(
        cost.evaluate(np.column_stack((interval_starts, interval_ends))), axis=1
    )
    total_cost += np.sum(interval_costs)

    return total_cost


def old_run_pelt(
    cost: BaseCost,
    penalty: float,
    min_segment_length: int,
    split_cost: float = 0.0,
    prune: bool = True,
    pruning_margin: float = 0.0,
) -> PELTResult:
    """Run the PELT algorithm.

    Currently agrees with the 'changepoint::cpt.mean' implementation of PELT in R.
    If the 'min_segment_length' is large enough to span more than a single changepoint,
    the algorithm can return a suboptimal partitioning.
    In that case, resort to the 'optimal_partitioning' algorithm.

    Parameters
    ----------
    cost: BaseCost
        The cost to use.
    penalty : float
        The penalty incurred for adding a changepoint.
    min_segment_length : int
        The minimum length of a segment, by default 1.
    split_cost : float, optional
        The cost of splitting a segment, to ensure that
        cost(X[t:p]) + cost(X[p:(s+1)]) + split_cost <= cost(X[t:(s+1)]),
        for all possible splits, 0 <= t < p < s <= len(X) - 1.
        By default set to 0.0, which is sufficient for
        log likelihood cost functions to satisfy the
        above inequality.
    prune: bool, optional
        If False, drop the pruning step, reverting to optimal partitioning.
        Can be useful for debugging and testing. By default set to True.
    pruning_margin : float, optional
        The percentage of pruning margin to use. By default set to zero.
        This is used to prune the admissible starts set.
        The pruning margin is used to avoid numerical issues when comparing
        the candidate optimal costs with the current optimal cost.

    Returns
    -------
    PELTResult
        Summary of the PELT algorithm run, containing:
        - `optimal_cost`: The optimal costs for each segment.
        - `previous_change_points`: The previous changepoints for each segment.
        - `pruning_fraction`: The fraction of pruning applied during the run.
        - `changepoints`: The final set of changepoints.
    """
    cost.check_is_fitted()
    n_samples = cost.n_samples
    min_segment_shift = min_segment_length - 1

    # Redefine Opt_cost[0] to start at 0.0, as done in 2014 PELT.
    # opt_cost = np.concatenate((np.array([0.0]), np.zeros(n_samples)))
    opt_cost = np.concatenate((np.array([-penalty]), np.zeros(n_samples)))

    # Cannot compute the cost for the first 'min_segment_shift' elements:
    opt_cost[1:min_segment_length] = np.inf

    # Compute the cost in [min_segment_length, 2*min_segment_length - 1] directly:
    non_changepoint_starts = np.zeros(min_segment_length, dtype=np.int64)
    non_changepoint_ends = np.arange(min_segment_length, 2 * min_segment_length)
    non_changepoint_intervals = np.column_stack(
        (non_changepoint_starts, non_changepoint_ends)
    )
    costs = cost.evaluate(non_changepoint_intervals)
    agg_costs = np.sum(costs, axis=1)
    opt_cost[min_segment_length : 2 * min_segment_length] = agg_costs

    # Aggregate number of cost evaluations:
    num_pelt_cost_evals = len(non_changepoint_starts)
    num_opt_part_cost_evals = len(non_changepoint_starts)

    # Store the previous changepoint for each latest start added.
    # Used to get the final set of changepoints after the loop.
    prev_cpts = np.repeat(0, n_samples)

    # Evolving set of admissible segment starts.
    cost_eval_starts = np.array(([0]), dtype=np.int64)

    observation_indices = np.arange(2 * min_segment_length - 1, n_samples).reshape(
        -1, 1
    )

    num_opt_part_cost_evals += int(
        (len(observation_indices) + 2) * (len(observation_indices) + 1) // 2 - 1
    )

    for current_obs_ind in observation_indices:
        latest_start = current_obs_ind - min_segment_shift
        opt_cost_obs_ind = current_obs_ind[0] + 1

        # Add the next start to the admissible starts set:
        cost_eval_starts = np.concatenate((cost_eval_starts, latest_start))
        cost_eval_ends = np.repeat(current_obs_ind + 1, len(cost_eval_starts))
        cost_eval_intervals = np.column_stack((cost_eval_starts, cost_eval_ends))
        costs = cost.evaluate(cost_eval_intervals)
        agg_costs = np.sum(costs, axis=1)

        num_pelt_cost_evals += len(cost_eval_starts)

        # Add the penalty for a new segment:
        candidate_opt_costs = opt_cost[cost_eval_starts] + agg_costs + penalty

        argmin_candidate_cost = np.argmin(candidate_opt_costs)
        opt_cost[opt_cost_obs_ind] = candidate_opt_costs[argmin_candidate_cost]
        prev_cpts[current_obs_ind] = cost_eval_starts[argmin_candidate_cost]

        if prune:
            # Trimming the admissible starts set: (reuse the array of optimal costs)
            current_obs_ind_opt_cost = opt_cost[opt_cost_obs_ind]

            abs_current_obs_opt_cost = np.abs(current_obs_ind_opt_cost)
            start_inclusion_threshold = (
                current_obs_ind_opt_cost
                # Apply pruning margin to the current optimal cost:
                + abs_current_obs_opt_cost * pruning_margin
                # Moved from 'negative' on left side
                # to 'positive' on right side.
                + penalty
                # Remove from right side of inequality.
                - split_cost
            )

            # Apply pruning:
            cost_eval_starts = cost_eval_starts[
                candidate_opt_costs <= start_inclusion_threshold
            ]

    pruning_fraction = 1.0 - num_pelt_cost_evals / num_opt_part_cost_evals

    pelt_result = PELTResult.new(
        optimal_costs=opt_cost[1:],
        previous_change_points=prev_cpts,
        pruning_fraction=pruning_fraction,
    )

    return pelt_result


def run_pelt_masked(
    cost: BaseCost,
    penalty: float,
    min_segment_length: int,
    split_cost: float = 0.0,
    pruning_margin: float = 0.0,
    allocation_multiplier: float = 5.0,  # Initial multiple of log(n_samples)
    growth_factor: float = 2.0,  # Geometric growth factor
) -> tuple[np.ndarray, list]:
    """Run the PELT algorithm.

    Currently agrees with the 'changepoint::cpt.mean' implementation of PELT in R.
    If the 'min_segment_length' is large enough to span more than a single changepoint,
    the algorithm can return a suboptimal partitioning.
    In that case, resort to the 'optimal_partitioning' algorithm.

    # Noteworthy: Uses pre-allocated arrays as much as possible,
    # to avoid excessive memory allocation and copying.

    Parameters
    ----------
    cost: BaseCost
        The cost to use.
    penalty : float
        The penalty incurred for adding a changepoint.
    min_segment_length : int
        The minimum length of a segment, by default 1.
    split_cost : float, optional
        The cost of splitting a segment, to ensure that
        cost(X[t:p]) + cost(X[p:(s+1)]) + split_cost <= cost(X[t:(s+1)]),
        for all possible splits, 0 <= t < p < s <= len(X) - 1.
        By default set to 0.0, which is sufficient for
        log likelihood cost functions to satisfy the
        above inequality.
    pruning_margin : float, optional
        The pruning margin to use. By default set to zero.
        This is used to prune the admissible starts set.
        The pruning margin is used to avoid numerical issues when comparing
        the candidate optimal costs with the current optimal cost.
    initial_capacity : float, optional
        The initial capacity of the pre-allocated arrays.
        This is a multiple of log(n_samples). Default is 5.0.
    growth_factor : float, optional
        The factor by which to grow the arrays when they need to be resized.
        Default is 2.0.

    Returns
    -------
    tuple[np.ndarray, list, float]
        The optimal costs, the changepoints, and cost evaluation time.
    """
    cost.check_is_fitted()
    n_samples = cost.n_samples
    min_segment_shift = min_segment_length - 1

    # Explicitly set the first element to 0.
    # Define "opt_cost[0]"" to start at 0.0, as done in 2014 PELT.
    opt_cost = np.concatenate((np.array([0.0]), np.zeros(n_samples)))

    # Cannot compute the cost for the first 'min_segment_shift' elements:
    opt_cost[1:min_segment_length] = 0.0

    # Compute the cost in [min_segment_length, 2*min_segment_length - 1] directly:
    non_changepoint_starts = np.zeros(min_segment_length, dtype=np.int64)
    non_changepoint_ends = np.arange(min_segment_length, 2 * min_segment_length)
    non_changepoint_intervals = np.column_stack(
        (non_changepoint_starts, non_changepoint_ends)
    )
    costs = cost.evaluate(non_changepoint_intervals)
    agg_costs = np.sum(costs, axis=1)
    opt_cost[min_segment_length : 2 * min_segment_length] = agg_costs + penalty

    # Store the previous changepoint for each latest start added.
    # Used to get the final set of changepoints after the loop.
    prev_cpts = np.repeat(0, n_samples)

    # Initialize smaller arrays with a fraction of n_samples capacity
    initial_size = max(2, int(np.log(n_samples) * allocation_multiplier))

    # Pre-allocate arrays with initial capacity
    start_capacity = initial_size
    starts_buffer = np.zeros(start_capacity, dtype=np.int64)
    interval_capacity = initial_size
    interval_buffer = np.zeros((interval_capacity, 2), dtype=np.int64)

    # Initialize with the first valid start position (position 0)
    n_valid_starts = 1
    starts_buffer[0] = 0  # First valid start is at position 0
    cost_eval_time = 0.0

    for current_obs_ind in range(2 * min_segment_length - 1, n_samples):
        latest_start = current_obs_ind - min_segment_shift

        # Add the next start position to the admissible set:
        # First check if we need to grow the arrays
        if n_valid_starts + 1 > start_capacity:
            # Grow arrays geometrically
            new_capacity = int(start_capacity * growth_factor)
            new_starts_buffer = np.zeros(new_capacity, dtype=np.int64)
            new_starts_buffer[:n_valid_starts] = starts_buffer[:n_valid_starts]
            starts_buffer = new_starts_buffer
            start_capacity = new_capacity

            # Also grow the interval buffer
            new_interval_capacity = int(interval_capacity * growth_factor)
            new_interval_buffer = np.zeros((new_interval_capacity, 2), dtype=np.int64)
            new_interval_buffer[:interval_capacity] = interval_buffer[
                :interval_capacity
            ]
            interval_buffer = new_interval_buffer
            interval_capacity = new_interval_capacity

        # Add the latest start to the buffer of valid starts
        starts_buffer[n_valid_starts] = latest_start
        n_valid_starts += 1

        # Set up intervals for cost evaluation
        current_end = current_obs_ind + 1

        # Fill the interval buffer with current valid starts and the current end
        interval_buffer[:n_valid_starts, 0] = starts_buffer[:n_valid_starts]
        interval_buffer[:n_valid_starts, 1] = current_end

        # Evaluate costs:
        cost_eval_t0 = time.perf_counter()
        agg_costs = np.sum(cost.evaluate(interval_buffer[:n_valid_starts]), axis=1)
        cost_eval_t1 = time.perf_counter()
        cost_eval_time += cost_eval_t1 - cost_eval_t0

        # Add the cost and penalty for a new segment (since last changepoint)
        # Reusing the agg_costs array to store the candidate optimal costs.
        agg_costs[:] += penalty + opt_cost[starts_buffer[:n_valid_starts]]
        candidate_opt_costs = agg_costs

        # Find the optimal cost and previous changepoint
        argmin_candidate_cost = np.argmin(candidate_opt_costs)
        min_start_idx = starts_buffer[argmin_candidate_cost]
        opt_cost[current_obs_ind + 1] = candidate_opt_costs[argmin_candidate_cost]
        prev_cpts[current_obs_ind] = min_start_idx

        # Pruning: update valid starts to exclude positions that cannot be optimal
        current_obs_ind_opt_cost = opt_cost[current_obs_ind + 1]
        abs_current_obs_opt_cost = np.abs(current_obs_ind_opt_cost)

        # Calculate pruning threshold with margin
        start_inclusion_threshold = (
            (
                current_obs_ind_opt_cost
                + abs_current_obs_opt_cost * (pruning_margin / 100.0)
            )
            + penalty  # Pruning inequality does not include added penalty.
            - split_cost  # Remove from right side of inequality.
        )

        # Apply pruning by filtering valid starts:
        valid_starts_mask = candidate_opt_costs <= start_inclusion_threshold
        n_new_valid_starts = np.sum(valid_starts_mask)
        starts_buffer[:n_new_valid_starts] = starts_buffer[:n_valid_starts][
            valid_starts_mask
        ]
        n_valid_starts = n_new_valid_starts

    return opt_cost[1:], get_changepoints(prev_cpts), cost_eval_time


def test_run_optimal_partitioning(
    cost: BaseCost, penalty: float, min_segment_length: int = 1
):
    cost.fit(changepoint_data)
    pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=True,
    )

    # Assert monotonicity of costs:
    if min_segment_length == 1:
        assert np.all(np.diff(pelt_result.optimal_costs) >= 0)
    assert len(pelt_result.changepoints) == n_segments - 1
    assert pelt_result.changepoints.tolist() == [seg_len]


def test_run_pelt(cost: BaseCost, penalty: float, min_segment_length=1):
    cost.fit(changepoint_data)
    pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
    )

    assert np.all(np.diff(pelt_result.optimal_costs) >= 0)
    assert len(pelt_result.changepoints) == n_segments - 1
    assert pelt_result.changepoints.tolist() == [seg_len]


def test_pelt_with_and_without_pruning_is_the_same(
    cost: BaseCost, penalty: float, min_segment_length: int = 1
):
    cost.fit(changepoint_data)

    opt_part_results = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=False,
    )

    pelt_results = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
    )

    assert np.array_equal(opt_part_results.changepoints, pelt_results.changepoints)
    np.testing.assert_array_almost_equal(
        pelt_results.optimal_costs, opt_part_results.optimal_costs
    )


@pytest.mark.parametrize("min_segment_length", [1, 5, 10])
@pytest.mark.parametrize(
    "signal_end_index", list(range(20, len(alternating_sequence) + 1, 5))
)
def test_pelt_on_tricky_data(
    cost: BaseCost, penalty: float, min_segment_length: int, signal_end_index: int
):
    """
    Test PELT on a slightly more complex data set. There are
    change points every 20 samples, and the mean of the segments
    changes drastically. And the PELT implementation agrees with
    the optimal partitioning as long as the segment length is
    less than 20.
    """
    # Original "run_pelt" found 7 changepoints.
    pruning_margin = 0.0
    cost.fit(alternating_sequence[0:signal_end_index])
    pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        pruning_margin=pruning_margin,
    )
    opt_part_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=True,
    )

    assert np.array_equal(pelt_result.changepoints, opt_part_result.changepoints)
    np.testing.assert_almost_equal(
        pelt_result.optimal_costs[-1],
        pelt_partition_cost(
            alternating_sequence[0:signal_end_index],
            pelt_result.changepoints,
            cost,
            penalty=penalty,
        ),
        decimal=10,
        err_msg="PELT cost for final observation does not match partition cost.",
    )
    np.testing.assert_array_almost_equal(
        pelt_result.optimal_costs, opt_part_result.optimal_costs
    )


@pytest.mark.parametrize("min_segment_length", range(1, 20))
def test_pelt_min_segment_lengths(cost: BaseCost, penalty: float, min_segment_length):
    """
    Test PELT on a slightly more complex data set. There are
    change points every 20 samples, and the mean of the segments
    changes drastically. And the PELT implementation agrees with
    the optimal partitioning as long as the segment length is
    less than or equal to 20.

    Segment length of 30 works again...
    """
    cost.fit(alternating_sequence)
    pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
    )

    cost.fit(alternating_sequence)
    opt_part_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=False,
    )

    assert np.array_equal(pelt_result.changepoints, opt_part_result.changepoints)
    np.testing.assert_array_almost_equal(
        pelt_result.optimal_costs, opt_part_result.optimal_costs, decimal=10
    )


@pytest.mark.parametrize("min_segment_length", range(31, 40))
def test_high_min_segment_length(cost: BaseCost, penalty: float, min_segment_length):
    """
    For all these segment lengths, the improved PELT implementation
    finds the same changepoints as the optimal partitioning.
    """
    cost.fit(alternating_sequence)
    # The PELT implementation (no longer) fails to find the same
    # changepoints as the optimal partitioning for these segment lengths,
    # when the segment length is greater than 30 and the pruning margin is zero.
    pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
    )

    opt_part_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=False,
    )

    assert np.array_equal(pelt_result.changepoints, opt_part_result.changepoints)
    np.testing.assert_array_almost_equal(
        pelt_result.optimal_costs, opt_part_result.optimal_costs
    )


@pytest.mark.parametrize("min_segment_length", [25])
def test_pelt_agrees_with_opt_part_longer_min_segment_length(
    cost: BaseCost, penalty: float, min_segment_length
):
    """
    For all these segment lengths, the PELT implementation
    fails to find the same changepoints as the optimal partitioning.
    """
    cost.fit(long_alternating_sequence)
    pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        pruning_margin=0.0,
    )

    opt_part_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=False,
    )

    assert np.array_equal(pelt_result.changepoints, opt_part_result.changepoints)
    np.testing.assert_array_almost_equal(
        pelt_result.optimal_costs, opt_part_result.optimal_costs
    )


@pytest.mark.parametrize("min_segment_length", [33, 39])
def test_comparing_skchange_to_ruptures_pelt_where_it_works(
    cost: BaseCost, penalty: float, min_segment_length: int
):
    """
    Test PELT on a slightly more complex data set. There are
    change points every n samples, and the mean of the segments
    changes drastically. And the PELT implementation agrees with
    the optimal partitioning as long as the segment length is
    less than or equal to 30.
    """
    cost.fit(long_alternating_sequence)

    opt_part_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=False,
    )
    opt_part_min_value = (
        evaluate_segmentation(cost, opt_part_result.changepoints)
        + (len(opt_part_result.changepoints)) * penalty
    )

    skchange_pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
    )
    skchange_pelt_min_value = (
        evaluate_segmentation(cost, skchange_pelt_result.changepoints)
        + (len(skchange_pelt_result.changepoints)) * penalty
    )

    rpt_model = rpt.Dynp(model="l2", min_size=min_segment_length, jump=1)
    rpt_model.fit(long_alternating_sequence)
    dyn_rpt_num_opt_part_cpts = np.array(
        rpt_model.predict(n_bkps=len(opt_part_result.changepoints))[:-1]
    )
    dyn_num_opt_part_cpts_min_value = (
        evaluate_segmentation(cost, dyn_rpt_num_opt_part_cpts)
        + (len(dyn_rpt_num_opt_part_cpts)) * penalty
    )

    ruptures_pelt_cpts = np.array(
        rpt.Pelt(model="l2", min_size=min_segment_length, jump=1).fit_predict(
            long_alternating_sequence, pen=penalty
        )[:-1]
    )
    rpt_pelt_min_value = (
        evaluate_segmentation(cost, ruptures_pelt_cpts)
        + (len(ruptures_pelt_cpts)) * penalty
    )
    assert np.array_equal(
        skchange_pelt_result.changepoints, opt_part_result.changepoints
    )
    assert np.array_equal(skchange_pelt_result.changepoints, dyn_rpt_num_opt_part_cpts)
    assert np.array_equal(skchange_pelt_result.changepoints, ruptures_pelt_cpts)

    assert skchange_pelt_min_value == opt_part_min_value
    assert skchange_pelt_min_value == dyn_num_opt_part_cpts_min_value
    assert skchange_pelt_min_value == rpt_pelt_min_value
    assert (
        np.abs(
            skchange_pelt_result.optimal_costs[-1] - opt_part_result.optimal_costs[-1]
        )
        < 1e-16
    )


@pytest.mark.parametrize("min_segment_length", [31, 32])
def test_compare_with_ruptures_pelt_where_restricted_pruning_works(
    cost: BaseCost, penalty: float, min_segment_length: int
):
    """
    Test PELT on a slightly more complex data set. There are
    change points every n samples, and the mean of the segments
    changes drastically. And the PELT implementation agrees with
    the optimal partitioning as long as the segment length is
    less than or equal to 30.
    """
    cost.fit(long_alternating_sequence)

    opt_part_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
        prune=False,
    )
    opt_part_min_value = (
        evaluate_segmentation(cost, opt_part_result.changepoints)
        + (len(opt_part_result.changepoints)) * penalty
    )

    # Compare with 'improved PELT':
    skchange_pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
    )
    skchange_pelt_min_value = (
        evaluate_segmentation(cost, skchange_pelt_result.changepoints)
        + (len(skchange_pelt_result.changepoints)) * penalty
    )

    rpt_model = rpt.Dynp(model="l2", min_size=min_segment_length, jump=1)
    rpt_model.fit(long_alternating_sequence)
    dyn_rpt_num_opt_part_cpts = np.array(
        rpt_model.predict(n_bkps=len(opt_part_result.changepoints))[:-1]
    )
    dyn_num_opt_part_cpts_min_value = (
        evaluate_segmentation(cost, dyn_rpt_num_opt_part_cpts)
        + (len(dyn_rpt_num_opt_part_cpts)) * penalty
    )

    ruptures_pelt_cpts = np.array(
        rpt.Pelt(model="l2", min_size=min_segment_length, jump=1).fit_predict(
            long_alternating_sequence, pen=penalty
        )[:-1]
    )
    rpt_pelt_min_value = (
        evaluate_segmentation(cost, ruptures_pelt_cpts)
        + (len(ruptures_pelt_cpts)) * penalty
    )
    assert np.array_equal(
        skchange_pelt_result.changepoints, opt_part_result.changepoints
    )
    assert np.array_equal(skchange_pelt_result.changepoints, dyn_rpt_num_opt_part_cpts)

    assert skchange_pelt_min_value == opt_part_min_value
    assert skchange_pelt_min_value == dyn_num_opt_part_cpts_min_value
    assert skchange_pelt_min_value < rpt_pelt_min_value
    assert (
        np.abs(
            skchange_pelt_result.optimal_costs[-1] - opt_part_result.optimal_costs[-1]
        )
        < 1e-16
    )


@pytest.mark.parametrize("min_segment_length", [1, 2, 5, 10])
def test_pelt_dense_changepoints_parametrized(cost: BaseCost, min_segment_length):
    """
    Test PELT with penalty=0.0 to ensure we get changepoints as dense as possible
    allowed by min_segment_length, for different min_segment_length values.
    """
    increasing_data = np.linspace(0, 1 * seg_len, seg_len).reshape(-1, 1)
    penalty = 0.0
    cost.fit(increasing_data)
    pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=min_segment_length,
    )

    # Expected changepoints are at every min_segment_length interval
    expected_changepoints = [
        i * min_segment_length
        for i in range(1, len(increasing_data) // min_segment_length)
    ]

    assert np.array_equal(pelt_result.changepoints, expected_changepoints)


def test_invalid_costs():
    """
    Test that PELT raises an error when given an invalid cost argument.
    """
    with pytest.raises(ValueError, match="cost"):
        PELT(cost="l2")
    with pytest.raises(ValueError, match="cost"):
        PELT(cost=CUSUM())
    with pytest.raises(ValueError, match="cost"):
        cost = L2Cost()
        cost.set_tags(is_penalised=True)  # Simulate a penalised score
        PELT(cost=cost)


@pytest.mark.parametrize("step_size", [3, 5, 10])
def test_pelt_with_step_size(cost: BaseCost, penalty: float, step_size: int):
    """Test PELT with jump parameter enabled and min_segment_length > 2."""

    # Run PELT with step_size > 1:
    jump_pelt_model = PELT(
        cost=cost,
        step_size=step_size,
        penalty=penalty,
    )
    jump_pelt_model.fit(alternating_sequence)
    pelt_changepoints = jump_pelt_model.predict(alternating_sequence)[
        "ilocs"
    ].to_numpy()

    assert np.all(pelt_changepoints % step_size == 0)

    # Compare with ruptures implementation for validation
    rpt_model = rpt.Pelt(model="l2", min_size=step_size, jump=step_size)
    rpt_changepoints = np.array(
        rpt_model.fit_predict(alternating_sequence, pen=penalty)[:-1]
    )

    # Check if our implementation finds reasonable changepoints:
    assert len(pelt_changepoints) > 0

    assert len(pelt_changepoints) == len(rpt_changepoints)
    assert np.array_equal(pelt_changepoints, rpt_changepoints)

    # Test that the optimal cost at the last observation is correct:
    pelt_cost = evaluate_segmentation(jump_pelt_model.fitted_cost, pelt_changepoints)
    pelt_min_value = pelt_cost + (len(pelt_changepoints)) * penalty
    assert np.isclose(
        jump_pelt_model.scores.to_numpy()[-1], pelt_min_value, atol=1e-10
    ), (
        f"Expected PELT cost at the last observation to be {pelt_min_value}, "
        f"got {jump_pelt_model.scores.to_numpy()[-1]}"
    )


def test_jump_pelt_pruning_fraction(cost: BaseCost, penalty: float):
    """Test pruning fraction is zero when not pruning, and < 1.0 when pruning."""
    cost.fit(alternating_sequence)
    step_size = 5

    # Run with pruning disabled
    pelt_result_no_pruning = _run_pelt_with_step_size(
        cost=cost,
        penalty=penalty,
        step_size=step_size,
        prune=False,
    )

    # Run with pruning enabled
    pelt_result_with_pruning = _run_pelt_with_step_size(
        cost=cost,
        penalty=penalty,
        step_size=step_size,
    )

    # Check that pruning fraction is 1.0 when pruning is disabled
    assert pelt_result_no_pruning.pruning_fraction == 0.0, (
        f"Expected pruning fraction to be 0.0 when prune=False, "
        f"got {pelt_result_no_pruning.pruning_fraction}"
    )

    # Check that pruning fraction is less than 1.0 when pruning is enabled
    assert 0.0 < pelt_result_with_pruning.pruning_fraction < 1.0, (
        f"Expected pruning fraction to be between 0.0 and 1.0 when prune=True, "
        f"got {pelt_result_with_pruning.pruning_fraction}"
    )

    # Also check that the changepoints are the same regardless of pruning
    assert np.array_equal(
        pelt_result_no_pruning.changepoints, pelt_result_with_pruning.changepoints
    ), "Changepoints should be the same regardless of pruning"

    # Check that optimal costs are the same
    np.testing.assert_array_equal(
        pelt_result_no_pruning.optimal_costs, pelt_result_with_pruning.optimal_costs
    )


@pytest.mark.parametrize("min_segment_length", [30, 35, 40])
@pytest.mark.parametrize("common_penalty", [1.6120892290743671, 1.6562305936619783])
def test_old_pelt_failing_with_large_min_segment_length(
    common_penalty: float, min_segment_length: int
):
    """Test the CROPS algorithm for path solutions to penalized CPD.

    Reference: https://arxiv.org/pdf/1412.3617
    """
    cost = L2Cost()

    # Generate test data:
    dataset = generate_alternating_data(
        n_segments=2,
        segment_length=100,
        p=1,
        mean=3.0,
        variance=4.0,
        random_state=42,
    )

    # Fit the change point detector:
    cost.fit(dataset)

    # Check that the results are as expected:
    # Optimal start for the final point: Index 101
    opt_part_result = _run_pelt(
        cost=cost,
        penalty=common_penalty,
        min_segment_length=min_segment_length,
        prune=False,
    )

    improved_pelt_result = _run_pelt(
        cost=cost,
        penalty=common_penalty,
        min_segment_length=min_segment_length,
    )

    old_pelt_result = old_run_pelt(
        cost=cost,
        penalty=common_penalty,
        min_segment_length=min_segment_length,
    )

    # PELT objective values:
    opt_part_min_value = evaluate_segmentation(
        cost, opt_part_result.changepoints
    ) + common_penalty * len(opt_part_result.changepoints)

    improved_pelt_min_value = evaluate_segmentation(
        cost, improved_pelt_result.changepoints
    ) + common_penalty * len(improved_pelt_result.changepoints)

    old_pelt_min_value = evaluate_segmentation(
        cost, old_pelt_result.changepoints
    ) + common_penalty * len(old_pelt_result.changepoints)

    # Compare results:
    assert np.array_equal(
        opt_part_result.changepoints, improved_pelt_result.changepoints
    )
    # Old PELT result should be different from the optimal partitioning result:
    assert not np.array_equal(
        opt_part_result.changepoints, old_pelt_result.changepoints
    )

    np.testing.assert_array_equal(
        opt_part_result.optimal_costs, improved_pelt_result.optimal_costs
    )
    assert old_pelt_min_value > opt_part_min_value, (
        f"Expected old PELT cost to be greater than optimal partitioning cost, "
        f"got {old_pelt_min_value} vs {opt_part_min_value}"
    )
    assert improved_pelt_min_value == opt_part_min_value, (
        f"Expected improved PELT cost to be equal to optimal partitioning cost, "
        f"got {improved_pelt_min_value} vs {opt_part_min_value}"
    )


def test_pelt_with_fewer_samples_than_min_segment_length_throws():
    """Test that PELT raises when `n_samples` is less than `min_segment_length`."""
    cost = L2Cost()
    data = np.random.randn(5, 1)  # Less than min_segment_length of 10
    cost.fit(data)

    with pytest.raises(
        ValueError,
        match="The `min_segment_length` cannot be larger than the number of samples",
    ):
        _run_pelt(cost, penalty=1.0, min_segment_length=10)


def test_jump_pelt_with_fewer_samples_than_step_size_throws():
    """Test that PELT-step_size  raises when `n_samples` is less than `step_size`."""
    cost = L2Cost()
    data = np.random.randn(5, 1)  # Less than step_size of 10

    # Creating the PELT model with step_size=10
    jump_pelt_model = PELT(cost=cost, step_size=10, penalty=1.0)

    with pytest.raises(
        ValueError,
        match="The `step_size` cannot be larger than the number of samples",
    ):
        jump_pelt_model.fit(data)
        jump_pelt_model.predict(data)


def test_jump_pelt_with_fewer_samples_than_twice_step_size_returns_single_interval():
    """Test that PELT-jump_pelt raises when `n_samples` is less than `step_size`."""
    cost = L2Cost()
    data = np.random.randn(10, 1)  # Less than step_size of 10

    # Creating the PELT model with step_size=6
    jump_pelt_model = PELT(cost=cost, step_size=6, penalty=1.0)

    jump_pelt_model.fit(data)
    jump_pelt_res = jump_pelt_model.predict(data)
    # Expect PELT-step_size to return an empty array when n_samples < 2 * step_size.
    np.testing.assert_array_equal(jump_pelt_res["ilocs"].to_numpy(), np.array([]))


def test_pelt_min_segment_length_one_agrees_with_regular_run_pelt(
    cost: BaseCost, penalty: float
):
    """Test that PELT with min_segment_length=1 agrees with run_pelt."""
    cost.fit(alternating_sequence)

    regular_pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=1,
    )
    no_pruning_pelt_result = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=1,
        prune=False,
    )
    np.testing.assert_array_equal(
        regular_pelt_result.changepoints, no_pruning_pelt_result.changepoints
    )
    np.testing.assert_array_equal(
        regular_pelt_result.optimal_costs, no_pruning_pelt_result.optimal_costs
    )

    min_seg_length_one_pelt_result = _run_pelt_min_segment_length_one(
        cost,
        penalty=penalty,
    )
    no_pruning_min_seg_length_one_pelt_result = _run_pelt_min_segment_length_one(
        cost,
        penalty=penalty,
        prune=False,
    )
    assert no_pruning_min_seg_length_one_pelt_result.pruning_fraction == 0.0, (
        "Expected no pruning when min_segment_length=1 and prune=False, "
        f"got {no_pruning_min_seg_length_one_pelt_result.pruning_fraction}"
    )

    assert (
        min_seg_length_one_pelt_result == regular_pelt_result
    ), "Expected PELT with min_segment_length=1 to agree with regular PELT."
    assert no_pruning_min_seg_length_one_pelt_result == no_pruning_pelt_result, (
        "Expected PELT with min_segment_length=1 and prune=False to agree with "
        "regular PELT with prune=False."
    )


def test_pelt_min_segment_length_one_throws_if_zero_sample():
    """Test that PELT with min_segment_length=1 raises an error if no samples."""
    cost = L2Cost()
    data = np.empty((0, 1))  # No samples
    cost.fit(data)

    with pytest.raises(
        ValueError,
        match="The number of samples for the fitted cost must be at least one.",
    ):
        _run_pelt_min_segment_length_one(cost, penalty=1.0)


def test_constructing_PELTResult_with_differing_array_sizes_raises_error():
    """Test that PELTResult raises an error if arrays have different sizes."""
    with pytest.raises(ValueError, match="All input arrays must have the same length."):
        PELTResult.new(
            optimal_costs=np.array([1.0, 2.0]),
            previous_change_points=np.array([0]),
            pruning_fraction=0.0,
        )


def test_comparing_PELTResult_with_non_PELTResult_returns_false():
    """Test that PELTResult comparison with non-PELTResult returns False."""
    pelt_result = PELTResult.new(
        optimal_costs=np.array([1.0, 2.0]),
        previous_change_points=np.array([0, 1]),
        pruning_fraction=0.0,
    )
    assert (
        not pelt_result == "not a PELTResult object"
    ), "Expected comparison with non-PELTResult to return False."


def test_PELTResult_cannot_be_hashed():
    """Test that PELTResult cannot be used as a key in a dictionary."""
    pelt_result = PELTResult.new(
        optimal_costs=np.array([1.0, 2.0]),
        previous_change_points=np.array([0, 1]),
        pruning_fraction=0.0,
    )
    with pytest.raises(TypeError, match="unhashable type: 'PELTResult'"):
        _ = {pelt_result: "value"}  # Attempt to use PELTResult as a dict key


def test_PELT_step_size_less_than_min_segment_length():
    """Test that PELT with step_size < min_segment_length raises an error."""
    cost = L2Cost()
    data = np.random.randn(100, 1)  # Enough samples for testing
    cost.fit(data)

    with pytest.raises(
        ValueError,
        match=(
            re.escape(
                "PELT `min_segment_length`(=10) cannot be "
                "greater than the `step_size`(=5) > 1."
            )
        ),
    ):
        PELT(
            cost=cost,
            step_size=5,  # Jump step less than min_segment_length
            penalty=1.0,
            min_segment_length=10,  # Min segment length is larger
        ).fit_predict(data)


def test_run_pelt_with_pruning_margin_decreases_pruning_fraction(
    cost: BaseCost, penalty: float
):
    """Test that increasing pruning margin decreases pruning fraction."""
    cost.fit(alternating_sequence)

    # Run PELT with a small pruning margin
    pelt_result_no_margin = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=4,
        pruning_margin=0.0,  # Small margin
    )

    # Run PELT with a larger pruning margin
    pelt_result_large_margin = _run_pelt(
        cost,
        penalty=penalty,
        min_segment_length=4,
        pruning_margin=10.0,  # Larger margin
    )

    # Check that the pruning fraction decreases with larger margin
    assert (
        pelt_result_large_margin.pruning_fraction
        < pelt_result_no_margin.pruning_fraction
    ), (
        "Expected larger pruning margin to decrease pruning fraction, "
        f"got {pelt_result_large_margin.pruning_fraction} < "
        f"{pelt_result_no_margin.pruning_fraction}"
    )


def test_run_pelt_min_seglen_one_with_pruning_margin_decreases_pruning_fraction(
    cost: BaseCost, penalty: float
):
    """Test that increasing pruning margin decreases pruning fraction."""
    cost.fit(alternating_sequence)

    # Run PELT with a small pruning margin
    pelt_result_no_margin = _run_pelt_min_segment_length_one(
        cost,
        penalty=penalty,
        pruning_margin=0.0,  # Small margin
    )

    # Run PELT with a larger pruning margin
    pelt_result_large_margin = _run_pelt_min_segment_length_one(
        cost,
        penalty=penalty,
        pruning_margin=10.0,  # Larger margin
    )

    # Check that the pruning fraction decreases with larger margin
    assert (
        pelt_result_large_margin.pruning_fraction
        < pelt_result_no_margin.pruning_fraction
    ), (
        "Expected larger pruning margin to decrease pruning fraction, "
        f"got {pelt_result_large_margin.pruning_fraction} < "
        f"{pelt_result_no_margin.pruning_fraction}"
    )


def test_run_pelt_step_size_with_pruning_margin_decreases_pruning_fraction(
    cost: BaseCost, penalty: float
):
    """Test that increasing pruning margin decreases pruning fraction."""
    cost.fit(alternating_sequence)

    # Run PELT with a small pruning margin
    pelt_result_no_margin = _run_pelt_with_step_size(
        cost,
        step_size=5,
        penalty=penalty,
        pruning_margin=0.0,  # Small margin
    )

    # Run PELT with a larger pruning margin
    pelt_result_large_margin = _run_pelt_with_step_size(
        cost,
        step_size=5,
        penalty=penalty,
        pruning_margin=10.0,  # Larger margin
    )

    # Check that the pruning fraction decreases with larger margin
    assert (
        pelt_result_large_margin.pruning_fraction
        < pelt_result_no_margin.pruning_fraction
    ), (
        "Expected larger pruning margin to decrease pruning fraction, "
        f"got {pelt_result_large_margin.pruning_fraction} < "
        f"{pelt_result_no_margin.pruning_fraction}"
    )



import numpy as np
import pytest

from skchange.base import BaseIntervalScorer
from skchange.change_scores import CHANGE_SCORES, to_change_score
from skchange.compose.penalised_score import PenalisedScore
from skchange.costs import COSTS
from skchange.costs.tests.test_all_costs import create_fixed_cost_test_instance
from skchange.datasets import generate_alternating_data
from skchange.tests.test_all_interval_scorers import skip_if_no_test_data

COST_INSTANCES = [
    to_change_score(create_fixed_cost_test_instance(cost))
    for cost in COSTS
    if cost.create_test_instance().get_tag("supports_fixed_param")
]
SCORE_INSTANCES = [score.create_test_instance() for score in CHANGE_SCORES]
PENALISED_SCORE_INSTANCES = [
    PenalisedScore(score)
    for score in COST_INSTANCES + SCORE_INSTANCES
    if not score.get_tag("is_penalised")
]
ALL_SCORE_INSTANCES = COST_INSTANCES + SCORE_INSTANCES + PENALISED_SCORE_INSTANCES


@pytest.mark.parametrize("change_score", ALL_SCORE_INSTANCES)
def test_scores(change_score: BaseIntervalScorer):
    """Test all available changepoint scores."""
    skip_if_no_test_data(change_score)

    n = 50
    df = generate_alternating_data(n_segments=1, segment_length=n, p=2, random_state=5)
    change_score.fit(df)
    splits = np.arange(15, n - 15, dtype=int).reshape(-1, 1)
    cuts = np.column_stack(
        (np.zeros(splits.shape, dtype=int), splits, np.full(splits.shape, n, dtype=int))
    )
    scores = change_score.evaluate(cuts)

    score_is_penalised = change_score.get_tag("is_penalised")
    min_value = 0.0 if not score_is_penalised else -np.inf
    assert np.all(scores >= min_value)



import numpy as np
import pandas as pd
import pytest
from sktime.tests.test_all_estimators import VALID_ESTIMATOR_TAGS

from skchange.anomaly_scores import ANOMALY_SCORES
from skchange.base import BaseIntervalScorer
from skchange.change_scores import CHANGE_SCORES
from skchange.compose.penalised_score import PenalisedScore
from skchange.costs import COSTS
from skchange.datasets import generate_alternating_data, generate_anomalous_data

INTERVAL_SCORERS = COSTS + CHANGE_SCORES + ANOMALY_SCORES + [PenalisedScore]
VALID_SCORER_TAGS = list(VALID_ESTIMATOR_TAGS) + [
    "task",
    "distribution_type",
    "is_conditional",
    "is_aggregated",
    "is_penalised",
    "supports_fixed_param",
]


def skip_if_no_test_data(scorer: BaseIntervalScorer):
    distribution_type = scorer.get_tag("distribution_type")
    is_conditional = scorer.get_tag("is_conditional")
    if distribution_type == "Poisson" or is_conditional:
        pytest.skip(
            f"{scorer.__class__.__name__} does not have test data in place yet."
        )


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_task_tag_set(Scorer: type[BaseIntervalScorer]):
    scorer = Scorer.create_test_instance()
    valid_tasks = ["cost", "change_score", "saving", "local_anomaly_score"]
    assert scorer.get_tag("task") in valid_tasks


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_scorer_fit(Scorer: type[BaseIntervalScorer]):
    scorer = Scorer.create_test_instance()
    skip_if_no_test_data(scorer)
    x = generate_anomalous_data()
    x.index = pd.date_range(start="2020-01-01", periods=x.shape[0], freq="D")
    fit_scorer = scorer.fit(x)
    assert fit_scorer.is_fitted


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_scorer_evaluate(Scorer: type[BaseIntervalScorer]):
    scorer = Scorer.create_test_instance()
    skip_if_no_test_data(scorer)
    x = generate_anomalous_data()
    x.index = pd.date_range(start="2020-01-01", periods=x.shape[0], freq="D")
    scorer.fit(x)
    cut1 = np.linspace(0, 40, scorer._get_required_cut_size(), dtype=int)

    results = scorer.evaluate(cut1)
    assert isinstance(results, np.ndarray)
    assert results.ndim == 2
    assert len(results) == 1

    cut2 = np.linspace(10, 40, scorer._get_required_cut_size(), dtype=int)
    cuts = np.array([cut1, cut2])
    results = scorer.evaluate(cuts)
    assert isinstance(results, np.ndarray)
    assert results.ndim == 2
    assert len(results) == len(cuts)


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_scorer_evaluate_by_evaluation_type(Scorer: BaseIntervalScorer):
    scorer = Scorer.create_test_instance()
    skip_if_no_test_data(scorer)
    n_segments = 1
    seg_len = 50
    p = 3
    df = generate_alternating_data(
        n_segments=n_segments,
        mean=20,
        segment_length=seg_len,
        p=p,
        random_state=15,
    )

    scorer.fit(df)
    cut1 = np.linspace(0, 20, scorer._get_required_cut_size(), dtype=int)
    cut2 = np.linspace(20, 40, scorer._get_required_cut_size(), dtype=int)
    cuts = np.array([cut1, cut2])

    results = scorer.evaluate(cuts)

    is_aggregated = scorer.get_tag("is_aggregated")
    is_conditional = scorer.get_tag("is_conditional")
    if is_aggregated:
        assert results.shape == (2, 1)
    elif not is_aggregated and is_conditional:
        assert results.shape[0] == 2
        assert results.shape[1] >= 1 and results.shape[1] <= p - 1
    else:
        assert results.shape == (2, p)


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_scorer_invalid_cuts(Scorer: BaseIntervalScorer):
    scorer = Scorer.create_test_instance()
    skip_if_no_test_data(scorer)
    x = generate_anomalous_data()
    x.index = pd.date_range(start="2020-01-01", periods=x.shape[0], freq="D")
    scorer.fit(x)
    with pytest.raises(ValueError):
        cut = np.linspace(0, 10, scorer._get_required_cut_size(), dtype=float)
        scorer.evaluate(cut)
    with pytest.raises(ValueError):
        cut = np.linspace(0, 10, scorer._get_required_cut_size() - 1, dtype=int)
        scorer.evaluate(cut)
    with pytest.raises(ValueError):
        cut = np.linspace(0, 10, scorer._get_required_cut_size() + 1, dtype=int)
        scorer.evaluate(cut)
    with pytest.raises(ValueError):
        cut = np.linspace(10, 0, scorer._get_required_cut_size(), dtype=int)
        scorer.evaluate(cut)


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_scorer_min_size(Scorer: BaseIntervalScorer):
    scorer = Scorer.create_test_instance()
    assert scorer.min_size is None or scorer.min_size >= 1

    skip_if_no_test_data(scorer)
    x = generate_anomalous_data()
    scorer.fit(x)
    assert scorer.min_size >= 1


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_scorer_param_size(Scorer: BaseIntervalScorer):
    scorer = Scorer.create_test_instance()
    assert scorer.get_model_size(1) >= 0

    skip_if_no_test_data(scorer)
    x = generate_anomalous_data()
    scorer.fit(x)
    assert scorer.get_model_size(1) >= 0


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_valid_interval_scorer_class_tags(Scorer: type[BaseIntervalScorer]):
    """Check that Scorer class tags are in VALID_SCORER_TAGS."""
    for tag in Scorer.get_class_tags().keys():
        msg = "Found invalid tag: %s" % tag
        assert tag in VALID_SCORER_TAGS, msg


@pytest.mark.parametrize("Scorer", INTERVAL_SCORERS)
def test_valid_interval_scorer_tags(Scorer: type[BaseIntervalScorer]):
    """Check that Scorer tags are in VALID_SCORER_TAGS."""
    scorer = Scorer.create_test_instance()
    for tag in scorer.get_tags().keys():
        msg = "Found invalid tag: %s" % tag
        assert tag in VALID_SCORER_TAGS, msg
